#define NULL 0
#include <iostream.h>
#include <strings.h>

typedef long Int ;
typedef unsigned long UInt ;

#define NULLPTR NULL
#define NULLCHAR '\000'

inline bool isNullPtr(void* ptr) { return ptr == NULLPTR ; } ;


class myString {
   public:
      char* data ;
      UInt length ;
      bool originalp ;

      inline myString& _let(char *s) {
	 data = s ;
	 length = strlen(s) ;
	 originalp = false ;
	 return *this ;
      } ;
      inline myString& _let(myString s) { return _let(s.data) ; } ;

      inline myString& _copy(const char *s) {
	 length = strlen(s) ;
	 data = new char[length+1] ;
	 strcpy(data,s) ;
	 originalp = true ;
	 return *this ;
      } ;
      inline myString& _copy(const myString s) { return _copy(s.data) ; } ;

      inline myString() { data = NULL ; length = 0 ; originalp = false ; };
      inline myString(const char *dt) { this->_copy(dt) ; } ;
      inline myString(UInt size) {
	 length = size ;
	 data = new char[length] ;
	 originalp = true ;
      } ;
      inline myString& operator= (const myString& s) {
	 return this->_copy(s) ; 
      } ;
      inline myString& operator= (const char* s) {
	 return this->_copy(s) ;
      } ;
      inline operator char* () { return data ; } ;

      inline myString operator+ (const myString& s) {
	 return *this + s.data ;
      } ;
      inline myString operator+ (const char* s) const {
	 Int dlen = strlen(data) ;
	 Int slen = strlen(s) ;
	 Int newlen = dlen + slen + 1 ;
	 myString newstr = myString(newlen) ;
	 strncpy(newstr.data,data,dlen) ;
	 strncpy(&(newstr.data[length]),s,slen) ;
	 newstr.data[newlen-1] = NULLCHAR ;
	 return newstr ;
      } ; 

      inline friend ostream& operator<< (ostream& o, const myString& s) {
	 o << s.data ;
      } ;
};


#define DEFAULT_ARRAY_SIZE	0
#define DEFAULT_EXPAND_FACTOR	2
#define DEFAULT_MEMORY_NAME	"Memory"



//======================================================================
// AMemory<Content>
//	arrayed memory 
template <class _Contents_>
class AMemory {
   public:
      //--------------------------------------------------
      // LocalTypeDef

      typedef AMemory<_Contents_> amemory_t ; 
      typedef _Contents_ cont_t ;

      //--------------------------------------------------
      // Member Variables

      cont_t* table ;
      UInt _top ;
      UInt size ;
      UInt expand_factor ;
      char* name ;

      //--------------------------------------------------
      // Constructor

      inline AMemory<_Contents_>(char *nm = DEFAULT_MEMORY_NAME,
				 UInt sz = DEFAULT_ARRAY_SIZE,
				 UInt ef = DEFAULT_EXPAND_FACTOR)
      {
	 init(nm,sz,ef) ;
      } ;

      inline bool init(char* nm, UInt sz, UInt ef) {
	 name = nm ;
	 size = 0 ;
	 expand_factor = ef ;
	 _top = 0 ;
	 return expand(sz,false) ;
      } ;
      
      //--------------------------------------------------
      // Expand

      inline bool expand(bool verbosep = false) {
	 return expand(size * expand_factor, verbosep) ;
      } ;

      inline bool expand(UInt newsize, bool verbosep) {
	 if(verbosep) {
	    cerr << "Expanding " << *this << ": " << size << "->" << newsize 
	       << " ... " ;
	    cerr.flush() ;
	 } 
	 if(newsize == 0 || newsize < size) return false ;
	 cont_t* oldtable = table ;
	 table = new cont_t[newsize] ;
	 if(table == NULLPTR) {
	    cerr << endl ;
	    cerr << "Warning: can't expand the size of " << *this
		 << "." << endl ;
	    return false ;
	 } 
	 for(UInt i = 0 ; i < size ; i++) {
	    table[i] = oldtable[i] ;
	 } 
	 size = newsize ;
	 if(verbosep) {
	    cerr << "Expanding is done." << endl ;
	 }
	 delete oldtable ;
	 return true ;
      } ;

      //--------------------------------------------------
      // describe

      inline void describe(ostream& ostr, bool detailp = true) const {
	 ostr << "#" << name << "[" << this 
	      << ":size=" << size 
	      << ",top=" << _top 
	      << ",ef=" << expand_factor 
	      << ",table=" << table 
	      << "]" ;
	 if(detailp) {
	    ostr << endl ;
	    for(UInt i = 0 ; i < _top ; i++) {
	       ostr << "\t" << table[i] << endl ;
	    }
	 }
      } ;

      inline friend ostream& operator<< (ostream& o, const amemory_t& am) {
	 am.describe(o, false) ;
	 return o ;
      } ;
} ;
      

//======================================================================
// APMemory
//    arrayed pointered memory
template <class _Contents_>
class APMemory : public AMemory<_Contents_ *> {
   public:
      //--------------------------------------------------
      // LocalTypeDef

      typedef APMemory<_Contents_> apmemory_t ; 
      typedef _Contents_ cont_t ;

      //--------------------------------------------------
      // Member Variables
      
      //--------------------------------------------------
      // Constructor

      inline APMemory<_Contents_>(char *nm = DEFAULT_MEMORY_NAME,
				  UInt sz = DEFAULT_ARRAY_SIZE,
				  UInt ef = DEFAULT_EXPAND_FACTOR) 
      {
	 init(nm,sz,ef) ;
      } ;

      inline bool init(char *nm, UInt sz, UInt ef) {
	 ((amemory_t*)this)->init(nm, 0, ef) ;
	 return expand(sz,false) ; 
      } ;

      //--------------------------------------------------
      // Expand
      
      inline bool expand(bool verbosep = false) {
	 return expand(size * expand_factor, verbosep) ;
      } ;

      inline bool expand(UInt newsize, bool verbosep) {
	 UInt oldsize = size ;
	 UInt sizedif = newsize - oldsize ;

	 ((amemory_t*)this)->expand(newsize, verbosep) ;

	 cont_t* newmem = new cont_t[sizedif] ;
	 if(newmem == NULLPTR) {
	    cerr << endl ;
	    cerr << "Warning: can't allocate new memory for " << *this
		 << "." << endl ;
	    return false ;
	 } 
	 for(UInt i = 0 ; i < sizedif ; i++) {
	    table[i+oldsize] = &(newmem[i]) ;
	 }
	 if(verbosep) {
	    cerr << "  Allocating is done." << endl ;
	 }
	 return true ;
      } ;

      //--------------------------------------------------
      // describe

      inline void describe(ostream& ostr, bool detailp = true) const {
	 ((amemory_t*)this)->describe(ostr, false) ;
	 if(detailp) {
	    ostr << endl ;
	    for(UInt i = 0 ; i < _top ; i++) {
	       ostr << "\t" << *(table[i]) << endl ;
	    }
	 }
      } ;

      inline friend ostream& operator<< (ostream& o, const apmemory_t& am) {
	 am.describe(o, false) ;
	 return o ;
      } ;
} ;


//======================================================================
// AStack<Content>
//
template <class _Contents_>
class AStack : public AMemory<_Contents_>{
   public:
      //--------------------------------------------------
      // LocalTypeDef

      typedef AStack<_Contents_> astack_t ; 

      //--------------------------------------------------
      // Constructor

      inline AStack<_Contents_>(char *nm = DEFAULT_MEMORY_NAME,
				UInt sz = DEFAULT_ARRAY_SIZE,
				UInt ef = DEFAULT_EXPAND_FACTOR)
      {
	 init(nm,sz,ef) ;
      } ;

      //--------------------------------------------------
      // Push/Pop

      inline astack_t* push(const cont_t& val) {
	 if(_top >= size) expand() ;
	 table[_top] = val ;
	 _top++ ;
	 return this ;
      } ;

      inline cont_t pop() {
	 if(_top <= 0) {
	    cerr << "Error: pop operation hits stack bottom (" << *this
		 << ")." << endl ;
	    exit(1) ;
	 }
	 _top-- ;
	 return table[_top] ;
      } ;
      inline cont_t popN(UInt c) {
	 if(_top < c) {
	    cerr << "Error: pop operation hits stack bottom (" << *this
		 << ")." << endl ;
	    exit(1) ;
	 }
	 _top -= c ;
	 return table[_top] ;
      } ;

      //--------------------------------------------------
      // access to top stack

      inline cont_t top() const {
	 if(_top <= 0) {
	    cerr << "Error: top operation hits stack bottom (" << *this
		 << ")." << endl ;
	    exit(1) ;
	 }
	 return table[_top-1] ;
      } ;

      inline cont_t topN(UInt c) const {
	 if(_top < c) {
	    cerr << "Error: topN operation hits stack bottom (" << *this
		 << ")." << endl ;
	    exit(1) ;
	 }
	 return table[_top-c-1] ;
      } ;

      inline UInt toppos() const { return _top ; } ;
      inline UInt setpos(UInt newtop) { return _top = newtop ; } ;

      inline void describe(ostream& o, bool detailp = true) const {
	 ((amemory_t*)this)->describe(o, false) ;
	 if(detailp) {
	    o << endl ;
	    for(UInt i = 0 ; i < _top ; i++) {
	       o << "\t" << table[_top-i] << endl ;
	    }
	    o << "\t___bottom___" << endl ;
	 }
      } ;

      inline friend ostream& operator<< (ostream& o, const astack_t& st) {
	 st.describe(o, false) ;
	 return o ;
      } ;
} ;


//======================================================================
// AHeap<Content>
//
template <class _Contents_>
class AHeap : public APMemory<_Contents_> {
   public:
      //--------------------------------------------------
      // LocalTypeDef

      typedef AHeap<_Contents_> aheap_t ; 

      //--------------------------------------------------
      // Constructor

      inline AHeap<_Contents_>(char *nm = DEFAULT_MEMORY_NAME,
			       UInt sz = DEFAULT_ARRAY_SIZE,
			       UInt ef = DEFAULT_EXPAND_FACTOR)
      {
	 init(nm,sz,ef) ;
      } ;

      //--------------------------------------------------
      // get new entity

      inline cont_t* get() {
	 if(_top >= size) expand() ;
	 cont_t* r = table[_top] ;
	 r->init() ;
	 _top++ ;
	 return r ;
      } ;

      //--------------------------------------------------
      // access to top stack

      inline UInt toppos() { return _top ; } ;
     
      inline bool rewind(UInt pos) { _top = pos ; return true ; } ;

};


//======================================================================
// SpStack<Content>
//	Spagetti Stack

template <class _Contents_>
class SpStackEntry {
   public:
      //--------------------------------------------------
      // LocalTypeDef
      typedef SpStackEntry<_Contents_> ss_entry_t ;
      typedef _Contents_ cont_t ;

      //--------------------------------------------------
      // Member Variables

      cont_t entity ;
      ss_entry_t* next ;

      //--------------------------------------------------
      // Constructor
      
      inline SpStackEntry<_Contents_>(ss_entry_t* nx = NULLPTR) {
	 next = nx ; 
      } ;
      inline SpStackEntry<_Contents_>(cont_t& ent, ss_entry_t nx = NULLPTR) {
	 entity = ent ;
	 next = nx ;
      } ;

      //--------------------------------------------------
      // Initializer
      inline bool init() {
	 next = NULLPTR ;
      } ;	 

      //--------------------------------------------------
      // Describe

      inline void describe(ostream& o, bool detailp = true) const {
	 if(detailp) {
	    o << "#SpStackEntry[" << this << "]" << endl ; 
	    for(const ss_entry_t* p = this ; p != NULLPTR ; p = p->next ) {
	       o << "\t" << entity << endl ;
	    }
	 } else {
	    o << "#SpStackEntry[" << entity << "]" ;
	 }
      } ;

      inline friend ostream& operator<< (ostream& o, const ss_entry_t& ss) {
	 ss.describe(o, false) ;
	 return o ;
      } ;
} ;

template <class _Contents_>
class SpStack : public AHeap< SpStackEntry<_Contents_> >{
   public:
      typedef SpStack<_Contents_> spstack_t ;
      typedef SpStackEntry<_Contents_> ss_entry_t ;
      typedef _Contents_ cont_t ;

      //--------------------------------------------------
      // Member Variables

      ss_entry_t* stackTop ;

      //--------------------------------------------------
      // Constructor
      
      inline SpStack<_Contents_>(char *nm = DEFAULT_MEMORY_NAME,
				 UInt sz = DEFAULT_ARRAY_SIZE,
				 UInt ef = DEFAULT_EXPAND_FACTOR)
      {
	 init(nm, sz, ef) ;
      } ;

      inline bool init(char *nm, UInt sz, UInt ef) {
	 ((aheap_t*)this)->init(nm,sz,ef) ;
	 stackTop = NULLPTR ;
      } ;
	    
      //--------------------------------------------------
      // Push/Pop

      inline ss_entry_t* push(const cont_t& ent) {
	 ss_entry_t* newtop = get() ;
	 newtop->entity = ent ;
	 newtop->next = stackTop ;
	 stackTop = newtop ;
	 return newtop ;
      } ;

      inline ss_entry_t* _pop() {
	 ss_entry_t* oldtop = stackTop ;
	 stackTop = oldtop->next ;
	 return oldtop ;
      } ;

      inline cont_t& pop() {
	 return _pop()->entity ;
      } ;

      inline cont_t& top() const {
	 return stackTop->entity ;
      } ;

      inline cont_t& topN(UInt c) const {
	 ss_entry_t* ent = stackTop ;
	 for(UInt i = 0 ; i < c ; i++) {
	    if(ent == NULLPTR) break ;
	    ent = ent->next ;
	 }
	 if(ent == NULLPTR) {
	    cerr << "Error: topN operation hits stack bottom (" << *this
		 << ")." << endl ;
	    exit(1) ;
	 }
	 return ent->entity ;
      } ;

      inline void setStackTop(ss_entry_t* newtop) {
	 stackTop = newtop ;
      } ;

      //--------------------------------------------------
      // describe

      inline void describe(ostream& ostr, bool detailp = true) const {
	 ((aheap_t*)this)->describe(ostr, false) ;
	 if(detailp) {
	    ostr << endl ;
	    for(ss_entry_t* top = stackTop ; top != NULLPTR ; 
					     top = top->next){
	       ostr << "\t" << top->entity << endl ;
	    }
	    ostr << "\t___bottom___" << endl ;
	 } else {
	    ostr << "<" << stackTop << ">" ;
	 }
      } ;

      inline friend ostream& operator<< (ostream& o, const spstack_t& sps) {
	 sps.describe(o, false) ;
	 return o ;
      } ;
} ;

      
typedef SpStack<int> GDSpStack ;

GDSpStack GDS = GDSpStack("GDSpStack",100) ;

int main(int argc, char** argv) {
   cout << GDS << endl ;
   GDS.push(1) ;
   cout << GDS << endl ;
}

