// -*- Mode:C++ -*-
//Header:
//File: smach.h
//Title: stach machine
//Author: NODA Itsuki
//Date: 1998/4/1
//EndHeader:

//ModifyHistory:
// 1998/4/1: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#ifndef _smach_h_
#define _smach_h_
////////////////////////////////////////////////////////////////////////

//======================================================================
// Includes
//
#include <pthread.h>

#include "base.h"
#include "gdata.h"
#include "unify.h"
#include "astack.h"
#include "aheap.h"
#include "slist.h"

//======================================================================
// Stack Machine
//

typedef AStack<GData> GDataAStack ;

class GDataStack : public GDataAStack {
   public:

      inline GDataStack() : GDataAStack() { } ;

      inline Int find(const GData dt) const {
	 for(Int i = 0 ; i < _top ; i++) {
	    if(eq(dt,(const GData)topN(i))) return i ;
	 } 
	 return -1 ;
      } ;

      inline GData remove(UInt n) {
	 if(n >= _top || n < 0) return NullSymbol ;
	 GData r = table[n-1] ;
	 for(UInt i = _top - n ; i < _top ; i++) {
	    table[i-1] = table[i] ;
	 }
	 _top-- ;
	 return r ;
      } ;

      inline UInt remove(GData dt) {
	 UInt r = find(dt) ;
	 remove(r) ;
      } ;

      inline UInt popUntil(GData dt) {
	 UInt r = find(dt) ;
	 popN(r+1) ;
	 return r ;
      } ;

      inline bool push(const GData d) {
	 GDataAStack::push(d) ;
      } ;
      inline bool push(const GData d1, const GData d2) {
	 push(d2) ; push(d1) ; 
	 return true ;
      } ;
      inline bool push(const GData d1, const GData d2, const GData d3) {
	 push(d3) ; push(d2) ; push(d1) ;
	 return true ;
      } ;
      inline bool push(const GData d1, const GData d2, const GData d3, 
		       const GData d4) {
			  push(d4) ; push(d3) ; push(d2) ; push(d1) ;
			  return true ;
      } ;
      inline bool push(const GData d1, const GData d2, const GData d3, 
		       const GData d4, const GData d5) {
			  push(d5) ; push(d4) ; push(d3) ; push(d2) ; push(d1) ;
			  return true ;
      } ;
      inline bool push(const GData d1, const GData d2, const GData d3, 
		       const GData d4, const GData d5, const GData d6){
			  push(d6) ; push(d5) ; push(d4) ; push(d3) ; push(d2) ; push(d1) ;
			  return true ;
      } ;
      inline bool push(const GData d1, const GData d2, const GData d3, 
		       const GData d4, const GData d5, const GData d6,
		       const GData d7){
			  push(d7) ;
			  push(d6) ; push(d5) ; push(d4) ; push(d3) ; push(d2) ; push(d1) ;
			  return true ;
      } ;
      inline bool push(const GData d1, const GData d2, const GData d3, 
		       const GData d4, const GData d5, const GData d6,
		       const GData d7, const GData d8){
			  push(d8) ; push(d7) ;
			  push(d6) ; push(d5) ; push(d4) ; push(d3) ; push(d2) ; push(d1) ;
			  return true ;
      } ;
} ;

class Player ;

class StackMachine {
   public:
      GDataStack opStack ;
      GDataStack argStack ;
      GDataStack envStack ;
      GDataStack auxStack ;

      pthread_mutex_t envMutex ;

      Player* player ;

      inline StackMachine() { pthread_mutex_init(&envMutex,NULL) ; } ;
      inline envLock() { pthread_mutex_lock(&envMutex) ; } ;
      inline envUnLock() { pthread_mutex_unlock(&envMutex) ; } ;

      inline pushEnv(GData cell) {
	 envLock() ;
	 envStack.push(cell) ;
	 envUnLock() ;
      } ;
      inline repushEnv(GData cell) {
	 envLock() ;	
	 Int c = envStack.find(cell) ;
	 if(c < 0)  envStack.push(cell) ;
	 envUnLock() ;
      } ;
      inline removeEnv(GData cell) {
	 envLock() ;
	 envStack.remove(cell) ;
	 envUnLock() ;
      } ;

};


////////////////////////////////////////////////////////////////////////
#endif
