// -*- Mode:C++ -*-
//Header:
//File: base.h
//Author: NODA Itsuki
//Date: 1998/2/6
//EndHeader:

#ifndef _base_h_
#define _base_h_

#define UseMyStringClass

////////////////////////////////////////////////////////////////////////
#ifdef NULL
#undef NULL
#define NULL 0
#endif
#include <iostream.h>
#include <strings.h>
#ifdef THREADED
#include <pthread.h>
#endif

//======================================================================
// Number Data Typedef
//
typedef long Int ;
typedef unsigned long UInt ;
typedef double Float ;
typedef unsigned int ID ;

/*#define NULLPTR	(void *)0*/
/*#define NULLPTR NULL*/
#define NULLPTR 0
#define NULLCHAR '\000'

#define Min(x,y) ((x>y)?(y):(x))
#define Max(x,y) ((x>y)?(x):(y))

inline bool isNullPtr(const void* ptr) { return ptr == NULLPTR ; } ;

//======================================================================
// myString Class
//
#ifndef UseMyStringClass
#include <String.h>
#endif

#ifdef UseMyStringClass
class myString {
   public:
      char* data ;
      UInt length ;
      bool originalp ;

      inline myString& _let(char *s) {
	 data = s ;
	 length = strlen(s) ;
	 originalp = false ;
	 return *this ;
      } ;
      inline myString& _let(myString s) { return _let(s.data) ; } ;

      inline myString& _copy(const char *s) {
	 length = strlen(s) ;
	 data = new char[length+1] ;
	 strcpy(data,s) ;
	 originalp = true ;
	 return *this ;
      } ;
      inline myString& _copy(const myString s) { return _copy(s.data) ; } ;

      inline myString() { data = NULLPTR ; length = 0 ; originalp = false ; };
      inline myString(const char *dt) { this->_copy(dt) ; } ;
      inline myString(const char *dt, bool copyp) {
	 if(copyp) {
	    _copy(dt) ;
	 } else {
	    _let(dt) ;
	 }
      } ;
	    
//  inline myString(const myString& dt) { this->_copy(dt) ; } ;
      inline myString(UInt size) {
	 length = size ;
	 data = new char[length] ;
	 originalp = true ;
      } ;

      inline bool operator== (const myString& s) {
	 return !strcmp(data,s.data) ;
      } ;
      inline bool operator== (const char* s) {
	 return !strcmp(data,s) ;
      } ;
      inline myString& operator= (const myString& s) {
	 return this->_copy(s) ; 
      } ;
      inline myString& operator= (const char* s) {
	 return this->_copy(s) ;
      } ;
      inline operator char* () { return data ; } ;

      inline myString operator+ (const myString& s) {
	 return *this + s.data ;
      } ;
      inline myString operator+ (const char* s) const {
	 Int dlen = strlen(data) ;
	 Int slen = strlen(s) ;
	 Int newlen = dlen + slen + 1 ;
	 myString newstr = myString(newlen) ;
	 strncpy(newstr.data,data,dlen) ;
	 strncpy(&(newstr.data[length]),s,slen) ;
	 newstr.data[newlen-1] = NULLCHAR ;
	 return newstr ;
      } ; 

      inline friend ostream& operator<< (ostream& o, const myString& s) {
	 return o << s.data ;
      } ;
};
#define String myString
#endif

//======================================================================
// my Mutex
//

class GMutex {
   public:
#ifdef THREADED
      pthread_mutex_t mutex ;
#endif

      inline GMutex() {
#ifdef THREADED
	 pthread_mutex_init(&mutex,NULLPTR) ;
#endif
      } ;

      inline void lock() {
#ifdef THREADED
	 pthread_mutex_lock(&mutex) ;
#endif
      } ;

      inline void unlock() {
#ifdef THREADED
	 pthread_mutex_unlock(&mutex) ;
#endif
      } ;
} ;

class GProcess {
   public:
#ifdef THREADED
      pthread_t id ;
#else
      Int id ;
#endif

} ;



////////////////////////////////////////////////////////////////////////
#endif
