// -*- Mode:C++ -*-
//Header:
//File: bipred.cc
//Title: built-in predicate 
//Author: NODA Itsuki
//Date: 1998/6/19
//EndHeader:

//ModifyHistory:
// 1998/6/19: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#include <math.h>
#include "gem.h"
#include "bipred.h"

#include "bipredutil.h"

//======================================================================
// True
//

bool bip_true(Gem* gem) {
   GData cgoal = gem->currentGoal().data ;

   if(isGSymbol(cgoal)) {
      return bipSuccess(gem) ;
   } else {
      return bipFail(gem) ;
   }
} ;

//======================================================================
// And
//

bool bip_and(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   UInt cutlevel = gem->astate->cutOrLevel ;
   
   if(isGTerm(icgoal.data)) {
      gem->popAndState() ;
      for(UInt i = _TermArity(icgoal.data) ; i > 0 ; i--) {
	 gem->pushAndState(InstGData(_TermArgv(icgoal.data,i-1),
				     icgoal.binding),
			   cutlevel) ;
      }
      return bipSuccess0(gem) ;
   } else {
      return bipFail(gem) ;
   }
} ;

//======================================================================
// Cut
//

bool bip_cut(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   UInt cutlevel = gem->astate->cutOrLevel ;
   
   if(cutlevel >0) cutlevel-- ;

   gem->popAndState() ;
   gem->orStateStack.setpos(cutlevel) ;
   
   return bipSuccess0(gem) ;
} ;

//======================================================================
// Repeat
//

bool bip_repeat(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if(isGSymbol(cgoal)) {
      gem->pushOrState() ;
      return bipSuccess(gem) ;
   } else {
      return bipFail(gem) ;
   }
}

//======================================================================
// Eq
//

bool bip_eq(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if(chkArity(cgoal,2)) {
      InstGData x = _fetchedArgv(icgoal,0) ;
      InstGData y = _fetchedArgv(icgoal,1) ;
      if(Unify(x,y,&(gem->uiheap))) {
	 return bipSuccess(gem) ;
      }
   }
   return bipFail(gem) ;
}

//======================================================================
// Halt
//

bool bip_halt(Gem* gem) {
   exit(0) ;
} ;

//======================================================================
// Trace/Notrace
//

bool bip_trace(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if(isGSymbol(cgoal)) {
      gem->traceon(Gem::TRC_normal) ;
      return bipSuccess(gem) ;
   } else if (chkArity(cgoal,1)) {
      GData g = fetchedArgv(icgoal,0) ;
      if(hasGem(g)) {
	 _SymGem(g)->traceon(Gem::TRC_normal) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal GEM") ;
      }
   } else {
      return bipFail(gem) ;
   } 
}

bool bip_systrace(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if(isGSymbol(cgoal)) {
      gem->traceon(Gem::TRC_kernel) ;
      return bipSuccess(gem) ;
   } else if (chkArity(cgoal,1)) {
      GData g = fetchedArgv(icgoal,0) ;
      if(hasGem(g)) {
	 _SymGem(g)->traceon(Gem::TRC_kernel) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal GEM") ;
      }
   } else {
      return bipFail(gem) ;
   } 
}

bool bip_notrace(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if(isGSymbol(cgoal)) {
      gem->notrace() ;
      return bipSuccess(gem) ;
   } else if (chkArity(cgoal,1)) {
      GData g = fetchedArgv(icgoal,0) ;
      if(hasGem(g)) {
	 _SymGem(g)->notrace() ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal GEM") ;
      }
   } else {
      return bipFail(gem) ;
   } 

}

//======================================================================
// showState
//

bool bip_showstate(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if(isGSymbol(cgoal)) {
      gem->describe(cout,true) ;
      return bipSuccess(gem) ;
   } else if (chkArity(cgoal,1)) {
      GData g = fetchedArgv(icgoal,0) ;
      if(hasGem(g)) {
	 _SymGem(g)->describe(cout,true) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal GEM") ;
      }
   } else {
      return bipFail(gem) ;
   } 
}

//======================================================================
// Show GCell Stack
//

bool bip_showenv(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if(isGSymbol(cgoal)) {
      gem->gcell->showGCellStack(cout,true) ;
      return bipSuccess(gem) ;
   } else if (chkArity(cgoal,1)) {
      GData g = fetchedArgv(icgoal,0) ;
      if(hasGCell(g)) {
	 _SymGCell(g)->showGCellStack(cout,true) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal GCell") ;
      }
   } else {
      return bipFail(gem) ;
   } 
}

//======================================================================
// Load
//

bool bip_load(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if(chkArity(cgoal,1)) {
      GData fname = fetchedArgv(icgoal,0) ;
      if(isGString(fname)) {
	 gem->popAndState() ;
	 gem->load(*_StrVal(fname)) ;
	 return bipSuccess0(gem) ;
      }
   }

   return bipError(gem,"Illegal load parameter") ;
}

//======================================================================
// listing
//

bool bip_listing(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if(isGSymbol(cgoal)) {
      gem->gcell->listing(cout,true) ;
      return bipSuccess(gem) ;
   } else if (chkArity(cgoal,1)) {
      GData gcsym = fetchedArgv(icgoal,0);
      if(hasGCell(gcsym)) {
	 _SymGCell(gcsym)->listing(cout,true) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Invalid Cell Name") ;
      }
   } else {
      return bipFail(gem) ;
   }
}

bool bip_listsymbol(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if(isGSymbol(cgoal)) {
      showSymTab() ;
      return bipSuccess(gem) ;
   } else {
      return bipFail(gem) ;
   }
}

//======================================================================
// NewCell
//

bool bip_newcell(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      GData gcsym = fetchedArgv(icgoal,0);
      if(isGSymbol(gcsym) && !hasGCell(gcsym)){
	 new GCell(_SymVal(gcsym)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Cell already exists") ;
      }
   } else {
      return bipFail(gem) ;
   }
}

//======================================================================
// PushCell
//
bool bip_push(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      GData gcsym = fetchedArgv(icgoal,0);
      if(hasGCell(gcsym)) {
	 gem->gcell->pushGCell(_SymGCell(gcsym)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Invarid Cell") ;
      }
   } else if (chkArity(cgoal,2)) {
      GData gcsym = fetchedArgv(icgoal,0) ;
      GData dstcell = fetchedArgv(icgoal,1) ;
      if(hasGCell(gcsym) && hasGCell(dstcell)) {
	 _SymGCell(dstcell)->pushGCell(_SymGCell(gcsym)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Invarid Cell") ;
      }
   } else {
      return bipFail(gem) ;
   } ;
}
      
//======================================================================
// PopCell
//
bool bip_pop(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      InstGData gcvar = _fetchedArgv(icgoal,0);
      GCell* gc = gem->gcell->popGCell() ;
      if(unifyWith(gc->gsymbol,gcvar,gem)) {
	 return bipSuccess(gem) ;
      } else {
	 return bipFail(gem) ;
      }
   } else if (chkArity(cgoal,2)) {
      InstGData gcvar = _fetchedArgv(icgoal,0) ;
      GData dstcell = fetchedArgv(icgoal,1) ;
      if(hasGCell(dstcell)) {
	 GCell* gc = _SymGCell(dstcell)->popGCell() ;
	 if(unifyWith(gc->gsymbol,gcvar,gem)) {
	    return bipSuccess(gem) ;
	 } else {
	    return bipFail(gem) ;
	 }
      } else {
	 return bipError(gem,"Invarid Cell") ;
      }
   } else {
      return bipFail(gem) ;
   } 
}

//======================================================================
// RemoveCell
//
bool bip_remove_cell(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      GData gcsym = fetchedArgv(icgoal,0);
      if(hasGCell(gcsym)) {
	 UInt r = gem->gcell->removeGCell(_SymGCell(gcsym)) ;
	 if(r > 0) {
	    return bipSuccess(gem) ;
	 } else {
	    return bipFail(gem) ;
	 }
      } else {
	 return bipError(gem,"Invarid Cell") ;
      }
   } else if (chkArity(cgoal,2)) {
      GData gcsym = fetchedArgv(icgoal,0) ;
      GData dstcell = fetchedArgv(icgoal,1) ;
      if(hasGCell(gcsym) && hasGCell(dstcell)) {
	 UInt r = _SymGCell(dstcell)->removeGCell(_SymGCell(gcsym)) ;
	 if(r > 0) {
	    return bipSuccess(gem) ;
	 } else {
	    return bipFail(gem) ;
	 }
      } else {
	 return bipError(gem,"Invarid Cell") ;
      }
   } else {
      return bipFail(gem) ;
   } ;
}

//======================================================================
// Remove1Cell
//
bool bip_remove_a_cell(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      GData gcsym = fetchedArgv(icgoal,0);
      if(hasGCell(gcsym)) {
	 UInt r = gem->gcell->removeGCell(_SymGCell(gcsym)) ;
	 if(r > 0) {
	    return bipSuccess(gem) ;
	 } else {
	    return bipFail(gem) ;
	 }
      } else {
	 return bipError(gem,"Invarid Cell") ;
      }
   } else if (chkArity(cgoal,2)) {
      GData gcsym = fetchedArgv(icgoal,0) ;
      GData dstcell = fetchedArgv(icgoal,1) ;
      if(hasGCell(gcsym) && hasGCell(dstcell)) {
	 bool r = _SymGCell(dstcell)->remove1GCell(_SymGCell(gcsym)) ;
	 if(r) {
	    return bipSuccess(gem) ;
	 } else {
	    return bipFail(gem) ;
	 }
      } else {
	 return bipError(gem,"Invarid Cell") ;
      }
   } else {
      return bipFail(gem) ;
   } ;
}

//======================================================================
// Environment
//

bool unifyGCellListSub(const InstGData& gdt, GCell* gcell, UInt n, Gem* gem) ;

bool unifyGCellListBody(const InstGData& gdt, GCell* gcell, Gem* gem) {
   if(gcell->gcellStack.bottomp()) {
      return unifyWith(gcell->gsymbol,gdt,gem) ;
   } else {
      Binding* newbinding = gem->uiheap.newBinding() ;
      if(!unify(gdt,InstGData(ConsTerm,newbinding),&(gem->uiheap))) 
	 return false ;
      if(!unifyWith(gcell->gsymbol,InstGData(ConsCarVar,newbinding),gem))
	 return false ;
      return unifyGCellListSub(InstGData(ConsCdrVar,newbinding),gcell,0,gem) ;
   }
}

bool unifyGCellListSub(const InstGData& gdt, GCell* gcell, UInt n, Gem* gem) {
   if(gcell->gcellStack.toppos() <= n) {
      return unifyWith(NullSymbol,gdt,gem) ;
   } else {
      Binding* newbinding = gem->uiheap.newBinding() ;
      if(!unify(gdt,InstGData(ConsTerm,newbinding),&(gem->uiheap)))
	 return false ;
      if(!unifyGCellListBody(InstGData(ConsCarVar,newbinding), 
			     gcell->gcellStack.topN(n),gem))
	 return false ;
      return unifyGCellListSub(InstGData(ConsCdrVar,newbinding),
			       gcell,n+1,gem) ;
   }
}

bool unifyGCellList(InstGData& gdt, GCell* gcell, Gem* gem) {
   BindingState bstate(gdt.binding, &(gem->uiheap)) ;
   bool r = unifyGCellListBody(gdt,gcell,gem) ;
   if(r) {
      return true ;
   } else {
      rewindBindingState(bstate) ;
      return false ;
   } 
}

bool bip_environment(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if (chkArity(cgoal,1)) {
      InstGData gdt = _fetchedArgv(icgoal,0) ;
      if(unifyGCellList(gdt,gem->gcell,gem)) {
	 return bipSuccess(gem) ;
      }
   }
   return bipFail(gem) ;
}

bool bip_in_env(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if (chkArity(cgoal,1,2)) {
      GData cell = fetchedArgv(icgoal,0) ;
      if(!hasGCell(cell)) return bipError(gem,"Illegal cell") ;
      GCell* src = gem->gcell ;
      if(chkArity(cgoal,2)) {
	 GData scell = fetchedArgv(icgoal,1) ;
	 if(!hasGCell(scell)) return bipError(gem,"Illegal cell") ;
	 src = _SymGCell(scell) ;
      }
      if(src->isInGCellStack(_SymGCell(cell)))
	 return bipSuccess(gem) ;
      else
	 return bipFail(gem) ;
   }
   return bipFail(gem) ;
}

//======================================================================
// assert
//

enum BIPAssertMode {
   BIPAM_asserta,
   BIPAM_assertz
} ;

bool bip_assert_fact(Gem* gem, BIPAssertMode mode) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if (chkArity(cgoal,1,2)) {
      GData head = fetchedValueArgv(icgoal,0,gem) ;
      GCell* dstcell = gem->agcell() ;
      if(chkArity(cgoal,2)) {
	 GData gcl = fetchedArgv(icgoal,1) ;
	 if(hasGCell(gcl)) {
	    dstcell = _SymGCell(gcl) ;
	 } else {
	    return bipError(gem,"Illegal GCell") ;
	 }
      }
      if      (mode == BIPAM_asserta) 
	 dstcell->asserta(GClause(head,TrueSymbol)) ;
      else if (mode == BIPAM_assertz)
	 dstcell->assertz(GClause(head,TrueSymbol)) ;
      else
	 return bipError(gem,"Illegal Assert Mode") ;

      return bipSuccess(gem) ;
   }
   return bipFail(gem) ;
}

bool bip_assert_rule(Gem* gem, BIPAssertMode mode) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if (chkArity(cgoal,2,3)) {
      GData head = fetchedValueArgv(icgoal,0,gem) ;
      GData body = fetchedValueArgv(icgoal,1,gem) ;
      GCell* dstcell = gem->agcell() ;
      if(chkArity(cgoal,3)) {
	 GData gcl = fetchedArgv(icgoal,2) ;
	 if(hasGCell(gcl)) {
	    dstcell = _SymGCell(gcl) ;
	 } else {
	    return bipError(gem,"Illegal GCell") ;
	 }
      }
      if      (mode == BIPAM_asserta) 
	 dstcell->asserta(GClause(head,body)) ;
      else if (mode == BIPAM_assertz)
	 dstcell->assertz(GClause(head,body)) ;
      else
	 return bipError(gem,"Illegal Assert Mode") ;

      return bipSuccess(gem) ;
   }
   return bipFail(gem) ;
}

bool bip_asserta_fact(Gem* gem) {
   return bip_assert_fact(gem, BIPAM_asserta) ;
}

bool bip_assertz_fact(Gem* gem) {
   return bip_assert_fact(gem, BIPAM_assertz) ;
}

bool bip_asserta_rule(Gem* gem) {
   return bip_assert_rule(gem, BIPAM_asserta) ;
}

bool bip_assertz_rule(Gem* gem) {
   return bip_assert_rule(gem, BIPAM_assertz) ;
}

//======================================================================
// Write
//
bool bip_write(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      InstGData gdt = _fetchedArgv(icgoal,0);
      cout << fetchValue(gdt,&(gem->uiheap)) ;
      cout.flush() ;
      return bipSuccess(gem) ;
   } 
   return bipFail(gem) ;
}

bool bip_nl(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (isGSymbol(cgoal)) {
      cout << endl ;
      return bipSuccess(gem) ;
   } 
   return bipFail(gem) ;
}

//======================================================================
// Arithmetic
//

bool bip_inc(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,2)) {
      GData val = fetchedArgv(icgoal,0);
      InstGData var = _fetchedArgv(icgoal,1);
      if(isGNumber(val)) {
	 unifyWith(GInteger(IntVal(val)+1),var,gem) ;
	 return bipSuccess(gem) ;
      } 
   }
   return bipFail(gem) ;
}

bool bip_dec(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,2)) {
      GData val = fetchedArgv(icgoal,0);
      InstGData var = _fetchedArgv(icgoal,1);
      if(isGNumber(val)) {
	 unifyWith(GInteger(IntVal(val)-1),var,gem) ;
	 return bipSuccess(gem) ;
      } 
   }
   return bipFail(gem) ;
}


bool bip_comp_gt(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,2)) {
      GData x = fetchedArgv(icgoal,0);
      GData y = fetchedArgv(icgoal,1);
      if(isGNumber(x)&&isGNumber(y)) {
	 if(FltVal(x)>FltVal(y))
	    return bipSuccess(gem) ;
	 else
	    return bipFail(gem) ;
      } else {
	 return bipError(gem,"Not a Number") ;
      }
   }
   return bipFail(gem) ;
}

bool bip_comp_ge(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,2)) {
      GData x = fetchedArgv(icgoal,0);
      GData y = fetchedArgv(icgoal,1);
      if(isGNumber(x)&&isGNumber(y)) {
	 if(FltVal(x)>=FltVal(y))
	    return bipSuccess(gem) ;
	 else
	    return bipFail(gem) ;
      } else {
	 return bipError(gem,"Not a Number") ;
      }
   }
   return bipFail(gem) ;
}

bool bip_comp_lt(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,2)) {
      GData x = fetchedArgv(icgoal,0);
      GData y = fetchedArgv(icgoal,1);
      if(isGNumber(x)&&isGNumber(y)) {
	 if(FltVal(x)<FltVal(y))
	    return bipSuccess(gem) ;
	 else
	    return bipFail(gem) ;
      } else {
	 return bipError(gem,"Not a Number") ;
      }
   }
   return bipFail(gem) ;
}

bool bip_comp_le(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,2)) {
      GData x = fetchedArgv(icgoal,0);
      GData y = fetchedArgv(icgoal,1);
      if(isGNumber(x)&&isGNumber(y)) {
	 if(FltVal(x)<=FltVal(y))
	    return bipSuccess(gem) ;
	 else
	    return bipFail(gem) ;
      } else {
	 return bipError(gem,"Not a Number") ;
      }
   }
   return bipFail(gem) ;
}

//======================================================================
// Is
//

GData evalIsForm(InstGData form, Gem* gem) {
   GData f = form.data ;

   if(isGNumber(f)) return f ;
   if(isGSymbol(f)) {
      if(eq(f,A_PISymbol))  return GFloat(M_PI) ;
      else                return ErrorSymbol ;
   }
   /*	Add	*/
   if        (chkHeadArity(f,A_AddSymbol,2)) {
      GData arg0 = evalIsForm(_fetchedArgv(form,0),gem) ;
      GData arg1 = evalIsForm(_fetchedArgv(form,1),gem) ;
      if(!isGNumber(arg0) || !isGNumber(arg1)) 
	 return ErrorSymbol ;
      if(isGInteger(arg0) && isGInteger(arg1)) 
	 return GInteger(IntVal(arg0) + IntVal(arg1)) ;
      else
	 return GFloat(FltVal(arg0) + FltVal(arg1)) ;
   /*	Sub	*/
   } else if (chkHeadArity(f,A_SubSymbol,2)) {
      GData arg0 = evalIsForm(_fetchedArgv(form,0),gem) ;
      GData arg1 = evalIsForm(_fetchedArgv(form,1),gem) ;
      if(!isGNumber(arg0) || !isGNumber(arg1)) 
	 return ErrorSymbol ;
      if(isGInteger(arg0) && isGInteger(arg1)) 
	 return GInteger(IntVal(arg0) - IntVal(arg1)) ;
      else
	 return GFloat(FltVal(arg0) - FltVal(arg1)) ;
   /*	Mul	*/
   } else if (chkHeadArity(f,A_MulSymbol,2)) {
      GData arg0 = evalIsForm(_fetchedArgv(form,0),gem) ;
      GData arg1 = evalIsForm(_fetchedArgv(form,1),gem) ;
      if(!isGNumber(arg0) || !isGNumber(arg1)) 
	 return ErrorSymbol ;
      if(isGInteger(arg0) && isGInteger(arg1)) 
	 return GInteger(IntVal(arg0) * IntVal(arg1)) ;
      else
	 return GFloat(FltVal(arg0) * FltVal(arg1)) ;
   /*	Div	*/
   } else if (chkHeadArity(f,A_DivSymbol,2)) {
      GData arg0 = evalIsForm(_fetchedArgv(form,0),gem) ;
      GData arg1 = evalIsForm(_fetchedArgv(form,1),gem) ;
      if(!isGNumber(arg0) || !isGNumber(arg1)) 
	 return ErrorSymbol ;
      return GFloat(FltVal(arg0) / FltVal(arg1)) ;
   /*	ERROR	*/
   } else {
      return ErrorSymbol ;
   }
}

bool bip_is(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if (chkArity(cgoal,2)) {
      InstGData var = _fetchedArgv(icgoal,0) ;
      InstGData form = _fetchedArgv(icgoal,1) ;
      
      GData val = evalIsForm(form, gem) ;
      if(eq(val,ErrorSymbol)) {
	 return bipError(gem,"Illegal IS form") ;
      }
      if(unifyWith(val, var, gem)) {
	 return bipSuccess(gem) ;
      }
   }
   return bipFail(gem) ;
}

//======================================================================
// Timer
//

bool bip_ctime(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,2)) {
      InstGData var0 = _fetchedArgv(icgoal,0);
      InstGData var1 = _fetchedArgv(icgoal,1);
      struct timeval ctime ;
      gettimeofday(&ctime,NULL) ;

      if(unifyWith(GInteger(ctime.tv_sec),var0,gem) &&
	 unifyWith(GInteger(ctime.tv_usec),var1,gem)) {
	 return bipSuccess(gem) ;
      }
   } 
   return bipFail(gem) ;
}

bool bip_utimedif(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,5)) {
      GData stsec = fetchedArgv(icgoal,0);
      GData stusec = fetchedArgv(icgoal,1);
      GData edsec = fetchedArgv(icgoal,2);
      GData edusec = fetchedArgv(icgoal,3);
      InstGData var = _fetchedArgv(icgoal,4);

      Int dsec = IntVal(edsec) - IntVal(stsec) ;
      Int dusec = dsec * 1000000 + IntVal(edusec) - IntVal(stusec) ;
//      GData dif = GFloat((Float)dusec/1000000.0) ;
      GData dif = GInteger(dusec) ;

      if(unifyWith(dif,var,gem)) {
	 return bipSuccess(gem) ;
      }
   } 
   return bipFail(gem) ;
}

bool bip_sleep(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      GData sec = fetchedArgv(icgoal,0);
      if(isGInteger(sec)) {
	 Sleep(IntVal(sec)) ;
	 return bipSuccess(gem) ;
      }
   }
   return bipFail(gem) ;
}

bool bip_msleep(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      GData msec = fetchedArgv(icgoal,0);
      if(isGInteger(msec)) {
	 MSleep(IntVal(msec)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal Argument") ;
      }
   } else if (chkArity(cgoal,2)) {
      GData sec = fetchedArgv(icgoal,0) ;
      GData msec = fetchedArgv(icgoal,1) ;
      if(isGInteger(sec) && isGInteger(msec)) {
	 MSleep(IntVal(sec),IntVal(msec)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal Argument") ;
      }
   }
   return bipFail(gem) ;
}

bool bip_usleep(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      GData usec = fetchedArgv(icgoal,0);
      if(isGInteger(usec)) {
	 USleep(IntVal(usec)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal Argument") ;
      }
   } else if (chkArity(cgoal,2)) {
      GData sec = fetchedArgv(icgoal,0) ;
      GData usec = fetchedArgv(icgoal,1) ;
      if(isGInteger(sec) && isGInteger(usec)) {
	 USleep(IntVal(sec),IntVal(usec)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal Argument") ;
      }
   }
   return bipFail(gem) ;
}

bool bip_sleep_forever(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (isGSymbol(cgoal)) {
      SleepForever() ;
      return bipSuccess(gem) ;
   }
   return bipFail(gem) ;
}
   

//======================================================================
// Fork
//
bool bip_fork(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,2,3)) {
      GData newgoal = fetchedValueArgv(icgoal,0,gem) ;
      InstGData pid = _fetchedArgv(icgoal,1) ;
      Symbol* pname = NULLPTR ;
      if(chkArity(cgoal,3)) {
	 GData pnamesym = fetchedArgv(icgoal,2) ;
	 if(isGSymbol(pnamesym)) {
	    pname = _SymVal(pnamesym) ;
	 } else {
	    return bipError(gem,"Illegal Process Name") ;
	 }
      }
      Gem* newgem = new Gem(pname) ;
      newgem->gcell->copyGCellStack(gem->gcell) ;
      newgem->settopgoal(newgoal) ;
      pthread_create(&(newgem->pid),NULL,UnfoldLoop,newgem) ;
      unifyWith(GSymbol(newgem->name),pid,gem) ;
      
      return bipSuccess(gem) ;
   } else {
      return bipFail(gem) ;
   } 
}

//======================================================================
// PID
//

bool bip_pid(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      InstGData var = _fetchedArgv(icgoal,0);
      if(unifyWith(GSymbol(gem->name),var,gem)) {
	 return bipSuccess(gem) ;
      } else {
	 return bipFail(gem) ;
      }
   }
   return bipFail(gem) ;
}
      

//======================================================================
// CondBarrier
//

bool bip_newbarrier(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      InstGData var = _fetchedArgv(icgoal,0) ;
      CondBarrier* bar = new CondBarrier() ;
      if(unifyWith(GPointer((void*)bar),var,gem)) {
	 return bipSuccess(gem) ;
      } 
   }
   return bipFail(gem) ;
}

bool bip_barrier_wait(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      GData bar = fetchedArgv(icgoal,0) ;
      if(isGPointer(bar)) {
	 ((CondBarrier*)_PtrVal(bar))->lockwait() ;
	 return bipSuccess(gem) ;
      }
   }
   return bipFail(gem) ;
}

bool bip_barrier_signal(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1)) {
      GData bar = fetchedArgv(icgoal,0) ;
      if(isGPointer(bar)) {
	 ((CondBarrier*)_PtrVal(bar))->signal() ;
	 return bipSuccess(gem) ;
      }
   }
   return bipFail(gem) ;
}
      
//======================================================================
// GenSym
//

bool bip_gensym(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if (chkArity(cgoal,1,2)) {
      InstGData var = _fetchedArgv(icgoal,0) ;
      GData seed = GenSymSeedSymbol ;
      if(chkArity(cgoal,2)) 
	 seed = fetchedArgv(icgoal,1) ;

      if(!isGSymbol(seed))
	 return bipError(gem,"Illegal gensym seed") ;

      if(unifyWith(genGSymbol(seed),var,gem)) 
	 return bipSuccess(gem) ;
      else
	 return bipFail(gem) ;
   }

   return bipFail(gem) ;
}
   
//======================================================================
// GDataBox
//

bool bip_assign_gdbox(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if(chkArity(cgoal,1)) {
      GData sym = fetchedArgv(icgoal,0) ;
      if(isGSymbol(sym)) {
	 bool r = _SymVal(sym)->assignGDBox() ;
	 if(!r)
	    return bipError(gem,"Symbol already has a gdbox") ;
	 else
	    return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Not a Symbol") ;
      }
   }
   return bipFail(gem) ;
}

bool bip_gdbox_get(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if(chkArity(cgoal,4)) {
      GData sym = fetchedArgv(icgoal,0) ;
      InstGData var = _fetchedArgv(icgoal,1) ;
      GData rp = fetchedArgv(icgoal,2) ;
      GData wp = fetchedArgv(icgoal,3) ;

      if(!isGSymbol(sym)) 
	 return bipError(gem,"Not a Symbol") ;
      if(!hasGDBox(sym))
	 return bipError(gem,"Symbol does not have a gdbox") ;

      bool retractp ;
      if(eq(rp,YesSymbol)) 
	 retractp = true ;
      else if (eq(rp,NoSymbol))
	 retractp = false ;
      else 
	 return bipError(gem,"Illegal Retract Mode (yes/no)") ;
      
      bool waitp ;
      if(eq(wp,YesSymbol)) 
	 waitp = true ;
      else if (eq(wp,NoSymbol))
	 waitp = false ;
      else 
	 return bipError(gem,"Illegal Wait Mode (yes/no)") ;

      GData val ;
      bool r = _SymGDBox(sym)->get(val,retractp,waitp) ;
      if(r) {
	 if(unifyWith(val,var,gem)) 
	    return bipSuccess(gem) ;
	 else
	    return bipFail(gem) ;
      } else {
	 return bipFail(gem) ;
      }
   }
   return bipFail(gem) ;
}

bool bip_gdbox_put(Gem* gem) {
   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;

   if(chkArity(cgoal,3)) {
      GData sym = fetchedArgv(icgoal,0) ;
      GData val = fetchedArgv(icgoal,1) ;
      GData wp = fetchedArgv(icgoal,2) ;

      if(!isGSymbol(sym)) 
	 return bipError(gem,"Not a Symbol") ;
      if(!hasGDBox(sym))
	 return bipError(gem,"Symbol does not have a gdbox") ;

      GDataBox::PutType ptype ;
      if(eq(wp,YesSymbol)) 
	 ptype = GDataBox::PT_wait ;
      else if (eq(wp,NoSymbol))
	 ptype = GDataBox::PT_nowait ;
      else if (eq(wp,ForceSymbol))
	 ptype = GDataBox::PT_force ;
      else 
	 return bipError(gem,"Illegal Wait Mode (yes/no/force)") ;

      bool r = _SymGDBox(sym)->put(val,ptype) ;
      if(r) {
	 return bipSuccess(gem) ;
      } else {
	 return bipFail(gem) ;
      }
   }
   return bipFail(gem) ;
}

//======================================================================
// Data Type
//

//--------------------------------------------------
// var
//

bool bip_var(Gem* gem) {
   /*	$var(Var)	*/

   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if(!chkArity(cgoal,1)) return bipFail(gem) ;

   InstGData v = _fetchedArgv(icgoal,0) ;

   if(!isGround(v)) return bipSuccess(gem) ;
   else return bipFail(gem) ;
}
      
//--------------------------------------------------
// issymbol
//

bool bip_issymbol(Gem* gem) {
   /*	$issymbol(Val)	*/

   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if(!chkArity(cgoal,1)) return bipFail(gem) ;

   GData v = fetchedArgv(icgoal,0) ;

   if(isGSymbol(v)) return bipSuccess(gem) ;
   else return bipFail(gem) ;
}
      
//--------------------------------------------------
// isinteger
//

bool bip_isinteger(Gem* gem) {
   /*	$isinteger(Val)	*/

   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if(!chkArity(cgoal,1)) return bipFail(gem) ;

   GData v = fetchedArgv(icgoal,0) ;

   if(isGInteger(v)) return bipSuccess(gem) ;
   else return bipFail(gem) ;
}
      
//--------------------------------------------------
// isfloat
//

bool bip_isfloat(Gem* gem) {
   /*	$isfloat(Val)	*/

   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if(!chkArity(cgoal,1)) return bipFail(gem) ;

   GData v = fetchedArgv(icgoal,0) ;

   if(isGFloat(v)) return bipSuccess(gem) ;
   else return bipFail(gem) ;
}
      
//--------------------------------------------------
// isnumber
//

bool bip_isnumber(Gem* gem) {
   /*	$isnumber(Val)	*/

   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if(!chkArity(cgoal,1)) return bipFail(gem) ;

   GData v = fetchedArgv(icgoal,0) ;

   if(isGNumber(v)) return bipSuccess(gem) ;
   else return bipFail(gem) ;
}
      
//--------------------------------------------------
// isstring
//

bool bip_isstring(Gem* gem) {
   /*	$isstring(Val)	*/

   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if(!chkArity(cgoal,1)) return bipFail(gem) ;

   GData v = fetchedArgv(icgoal,0) ;

   if(isGString(v)) return bipSuccess(gem) ;
   else return bipFail(gem) ;
}

//--------------------------------------------------
// ispointer
//

bool bip_ispointerg(Gem* gem) {
   /*	$ispointer(Val)	*/

   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if(!chkArity(cgoal,1)) return bipFail(gem) ;

   GData v = fetchedArgv(icgoal,0) ;

   if(isGPointer(v)) return bipSuccess(gem) ;
   else return bipFail(gem) ;
}

//--------------------------------------------------
// term arity
//

bool bip_term_arity(Gem* gem) {
   /*	$term_arity(Term,Head,Arity)	*/

   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if(!chkArity(cgoal,3)) return bipFail(gem) ;

   GData v = fetchedArgv(icgoal,0) ;
   InstGData head = _fetchedArgv(icgoal,1) ;
   InstGData arity = _fetchedArgv(icgoal,2) ;

   if(!isGTerm(v)) return bipFail(gem) ;

   if(unifyWith(GSymbol(_TermHead(v)),head,gem) &&
      unifyWith(GInteger(_TermArity(v)),arity,gem))
      return bipSuccess(gem) ;
   else
      return bipFail(gem) ;
}
      
//--------------------------------------------------
// term argv
//

bool bip_term_argv(Gem* gem) {
   /*	$term_argv(Term,Argc,-Argv)	*/

   InstGData icgoal = gem->currentGoal() ;
   GData cgoal = icgoal.data ;
   
   if(!chkArity(cgoal,3)) return bipFail(gem) ;

   GData v = fetchedArgv(icgoal,0) ;
   GData argc = fetchedArgv(icgoal,1) ;
   InstGData argv = _fetchedArgv(icgoal,2) ;

   if(!isGTerm(v) || !isGNumber(argc)) return bipFail(gem) ;
   if(_TermArity(v) <= IntVal(argc)) return bipFail(gem) ;

   if(unifyWith(_TermArgv(v,IntVal(argc)),argv,gem))
      return bipSuccess(gem) ;
   else
      return bipFail(gem) ;
}
      

