//   -*- Mode: C++ -*-
//Header:
//EndHeader:

//Memo:
//  Number:
//    Integer:
//	Hex:  0xNNNN   N = [0-9a-fA-F]
//            0XNNNN
//      Dec:  sMNNN    S = [-|+|] M = [1-9], N = [0-9]
//	Oct:  0NNNN    N = [0-8]
//    Float:  sNNN.NNN[e|E]sNNN  (do not permit ".1")
//  Symbol:
//    Symbol = [a-z$][a-zA-Z$_]* | 'any string' | Operator
//    Operator = any symbol charactor except $ and _.
//  Variable:
//      
//EndMemo:

//======================================================================
// Include Files
//----------------------------------------------------------------------
#include "gbio.h"

//======================================================================
// Local Buffer
//----------------------------------------------------------------------

bool LocalBuffer::putChar(char c) {
   if(max > len) {
      buffer[len] = c ; 
      len++ ;
      return true ;
   } else {
      if(extendp) {
	 extendBuffer() ;
	 return putChar(c) ;
      } else {
	 return false ;
      }
   }
}
    

//======================================================================
// Utility Functions
//----------------------------------------------------------------------
Int skipSpace(istream& istr, ReadStatus& rstatus) {
   Int i=0 ;
   for(char c = istr.peek() ; isspace(c) ; c = istr.peek()) {
      getChar(istr,rstatus) ;
      i++ ;
   }
   return i ;
}

bool skipComment(istream& istr, ReadStatus& rstatus) {
   if(istr.peek() == COMMENTSTARTCHAR) {
      char c ;
      while((c = getChar(istr,rstatus)) != COMMENTENDCHAR) {}
      return true ;
   } else {
      return false ;
   }
}

void skipSpaceComment(istream& istr, ReadStatus& rstatus) {
   skipSpace(istr,rstatus) ;
   if(skipComment(istr,rstatus)) 
      skipSpaceComment(istr,rstatus) ;
}

bool readUntilChar(istream& istr, LocalBuffer& buf, 
		   char endchar, char escapechar,
		   bool discardlastp, ReadStatus& rstatus) 
{
   char c ;
   for(c = istr.peek() ; c != endchar ; c = istr.peek()) {
      if(c == EOF) {
	 rstatus.errorID = REI_IllegalEOF ;
	 return false ;
      } 
      c = getChar(istr,rstatus) ;
      if(c == escapechar) c = getChar(istr,rstatus) ;
      buf.putChar(c) ;
   }
   if(discardlastp) getChar(istr,rstatus) ;
   buf.putChar(EOSCHAR) ;
   return true ;
}

bool skipLine(istream& istr, ReadStatus& rstatus) {
   while(true) {
      if(istr.peek() == EOF) return false ;
      if(istr.peek() == NULLCHAR) return false ;
      if(CRCHAR == getChar(istr,rstatus)) return true ;
   }
}


//======================================================================
// readGData
//----------------------------------------------------------------------

GData readSentence(istream& istr, ReadStatus& rstatus) {
   rstatus.errorID = REI_None ;
   GData dt = readGData(istr, rstatus) ;
   if(rstatus.errorID == REI_None) {
      skipSpaceComment(istr,rstatus) ;
      char nextc = istr.peek() ;
      if(nextc == CLAUSEEND) getChar(istr,rstatus) ;
   }
   if(eq(dt,ErrorSymbol)) skipLine(istr,rstatus) ; 
   return dt ;
}
   

GData readGData(istream& istr, ReadStatus& rstatus) {
   skipSpaceComment(istr,rstatus) ;
   char nextc = istr.peek() ;
   GData dt ;
   switch(nextc) {
      case CLAUSEEND :
	 rstatus.errorID = REI_NilData ;
	 return ErrorSymbol;
      case QUOTESYMBOLBEGIN :
	 dt = readQuotedGSymbol(istr,rstatus) ;
	 break ;
      case STRINGBEGIN :
	 dt = readGString(istr,rstatus) ;
	 break ;
      default:
	 dt = readGNumberGSymbol(istr,rstatus) ;
	 break ;
   }
   if(isGSymbol(dt)) 
      dt = checkArgs(dt,istr,rstatus) ;

   return dt ;
}

GData readQuotedGSymbol(istream& istr, ReadStatus& rstatus) {
   if(istr.peek() != QUOTESYMBOLBEGIN) {
      rstatus.errorID = REI_IllegalSymbolFormat ;
      return ErrorSymbol ;
   }
   getChar(istr,rstatus) ;

   LocalBuffer buf(BUFSIZE) ;

   readUntilChar(istr,buf,QUOTESYMBOLEND,ESCAPECHAR,true,rstatus) ;

   return GSymbol(buf.buffer) ;
}

GData readGString(istream& istr, ReadStatus& rstatus) {
   if(istr.peek() != STRINGBEGIN) {
      rstatus.errorID = REI_IllegalStringFormat ;
      return ErrorSymbol ;
   }
   getChar(istr,rstatus) ;

   LocalBuffer buf(BUFSIZE) ;

   readUntilChar(istr,buf,STRINGEND,ESCAPECHAR,true,rstatus) ;

   return GString(buf.buffer) ;
}

GData readGNumberGSymbol(istream& istr, ReadStatus& rstatus) {
   LocalBuffer buf(BUFSIZE) ;

   char c = istr.peek() ;
   if(c == EOF) { 
      rstatus.errorID = REI_IllegalEOF ;
      return EOFSymbol ;
   }
   c = getChar(istr,rstatus) ;

   buf.putChar(c) ;
   if(isSignChar(c)) {				/* "+" or "-" */
      char nc = istr.peek() ;
      if(isdigit(nc)) {				/* "+[0-9]" */
	 return readGIntFloatRest(istr,buf,rstatus) ;	/* -> DecInt/Float */
      } else {
	 buf.finish() ;
	 return GSymbol(buf.buffer) ;			/* -> '+' or '-' */
      }
   } else if(isZeroChar(c)) {			/* "0" */
      char nc = istr.peek() ;
      if(isDotChar(c)) {				/* "0." */
	 return readGIntFloatRest(istr,buf,rstatus) ;	/* -> Float */
      } else if(isHexPreChar(nc)) {		/* "0x" or "0X" */
	 return readGHexIntRest(istr,buf,rstatus) ;	/* -> HexInt */
      } else {					/* "0" */
	 return readGOctIntRest(istr,buf,rstatus) ;	/* ->  OctInt */
      }
   } else if(isdigit(c)) {			/* "[1-9]" */
      return readGIntFloatRest(istr,buf,rstatus) ;	/* -> DecInt/Float */
   } else if(isSymHeadChar(c)) {			/* "[a-z$]" */
      return readGSymbolRest(istr,buf,rstatus) ;		/* -> Symbol */
   } else if(isVarHeadChar(c)) {			/* "[A-Z_]" */
      return readGVariableRest(istr,buf,rstatus);	      /* -> Variable */
   } else {					/* else */
      buf.finish() ;
      return GSymbol(buf.buffer) ;			/* -> Operator */
   }
}

GData readGIntFloatRest(istream& istr, LocalBuffer& buf, ReadStatus& rstatus){
   enum { IFM_beforePeriod, 
	  IFM_afterPeriod, 
	  IFM_afterExponent 
   } mode ;

   mode = IFM_beforePeriod ;
   while(true) {
      char c = istr.peek() ; 
      if (isdigit(c)) {
	 buf.putChar(getChar(istr,rstatus)) ;
      } else if (isDotChar(c)) {
	 if(mode == IFM_beforePeriod) {
	    buf.putChar(getChar(istr,rstatus)) ;
	    mode = IFM_afterPeriod ;
	 } else {
	    break ;
	 }
      } else if (isExpSepChar(c)) {
	 if(mode == IFM_afterExponent) {
	    break;
	 } else {
	    buf.putChar(getChar(istr,rstatus)) ;
	    char nc = istr.peek() ;
	    if(isSignChar(nc)) {
	       buf.putChar(getChar(istr,rstatus)) ;
	       nc = istr.peek() ;
	    }
	    if(!isdigit(nc)) {
	       rstatus.errorID = REI_IllegalNumberFormat ;
	       return ErrorSymbol ;
	    } 
	 }
      } else {
	 break ;
      }
   }
   buf.finish() ;
  
   if(mode == IFM_beforePeriod) {
      Int ival = strtol(buf.buffer,NULL,10) ;
      return GInteger(ival) ;
   } else {
      Float fval = strtod(buf.buffer,NULL) ;
      return GFloat(fval) ;
   } 
}

GData readGOctIntRest(istream& istr, LocalBuffer& buf, ReadStatus& rstatus){
   while(true) {
      char c = istr.peek() ;
      if (isOctIntChar(c)) {
	 buf.putChar(getChar(istr,rstatus)) ;
      } else {
	 break ;
      }
   }
   buf.finish() ;
  
   Int ival = strtol(buf.buffer,NULL,8) ;
   return GInteger(ival) ;
}

GData readGHexIntRest(istream& istr, LocalBuffer& buf, ReadStatus& rstatus){
   buf.putChar(getChar(istr,rstatus)) ;
   while(true) {
      char c = istr.peek() ;
      if(isxdigit(c)) {
	 buf.putChar(getChar(istr,rstatus)) ;
      } else {
	 break ;
      }
   }
   buf.finish() ;
  
   Int ival = strtol(buf.buffer,NULL,16) ;
   return GInteger(ival) ;
}

GData readGSymbolRest(istream& istr, LocalBuffer& buf, ReadStatus& rstatus){
   while(true) {
      char c = istr.peek() ;
      if (isSymBodyChar(c)){
	 buf.putChar(getChar(istr,rstatus)) ;
      } else {
	 break ;
      }
   }
   buf.finish() ;
  
   return GSymbol(buf.buffer) ;
}

GData readGVariableRest(istream& istr, LocalBuffer& buf, ReadStatus& rstatus){
   while(true) {
      char c = istr.peek() ;
      if(isSymBodyChar(c)) {
	 buf.putChar(getChar(istr,rstatus)) ;
      } else {
	 break ;
      }
   }
   buf.finish() ;
  
   return GVariable(buf.buffer) ;
}

GData checkArgs(const GData& head, istream& istr, ReadStatus& rstatus) {
   skipSpaceComment(istr,rstatus) ;
   char nextc = istr.peek() ;
   if(isArgBeginChar(nextc)) {
      getChar(istr,rstatus) ;
      return readArgs(head,istr,rstatus,0) ;
   } else {
      return head ;
   }
}

GData readArgs(const GData& head, istream& istr, ReadStatus& rstatus,
	       Int argn) 
{
   skipSpaceComment(istr,rstatus) ;
   char nextc = istr.peek() ;
   if(isArgEndChar(nextc)) {
      getChar(istr,rstatus) ;
      if(argn < 1)  {
	 rstatus.errorID = REI_IllegalTermFormat ;
	 return ErrorSymbol ;
      } else {
	 GData r = GTerm(head.val.symbol, argn) ;
	 return r ; 
      }
   } else if(isArgSepChar(nextc) && argn < 1) {
      rstatus.errorID = REI_IllegalTermFormat ;
      return ErrorSymbol ;
   } else if( (! isArgSepChar(nextc)) && argn > 0) {
      rstatus.errorID = REI_IllegalTermFormat ;
      return ErrorSymbol ;
   } else {
      if(isArgSepChar(nextc)) getChar(istr,rstatus) ;
      GData arg = readGData(istr,rstatus) ;
      GData r = readArgs(head,istr,rstatus,argn + 1) ;
      r.val.term->argv(argn) = arg ;
      return r ;
   } ;
}


bool GSymbol::shouldquotep() const {
   Int i = 0;
   for(char* c = val.symbol->name.data ; *c != NULLCHAR ; c++) {
      if(i == 0 && !isSymHeadChar(*c)) return true ;
      if(!isSymBodyChar(*c)) return true ;
      i++ ;
   }
   return false ;
}
    
//======================================================================
// output for debug
//----------------------------------------------------------------------
void dbp(GData dt) { cerr << dt ; }
void dbd(GData dt) { dt.describe(cerr) ; } 
