//   -*- Mode: C++ -*-
//Header:
//EndHeader:

#ifndef _gbio_h_
#define _gbio_h_
////////////////////////////////////////////////////////////////////////

//======================================================================
// Include Files
#include <stdlib.h>
#include <ctype.h>
#include <strstream.h>
#include "thread_util.h"
#include "gdata.h"


//======================================================================
// CHAR CONSTANTS

#define CRCHAR			'\n'
#define COMMENTSTARTCHAR	'%'
#define COMMENTENDCHAR  	'\n'
#define NULLCHAR		'\000'
#define EOSCHAR			NULLCHAR
#define ARGSEP			','
#define ARGBEGIN		'('
#define ARGEND			')'
#define CLAUSEEND		'.'
#define QUOTESYMBOLBEGIN	'\''
#define QUOTESYMBOLEND		'\''
#define STRINGBEGIN		'"'
#define STRINGEND		'"'
#define ESCAPECHAR		'\\'
#define PLUSCHAR		'+'
#define MINUSCHAR		'-'
#define ZEROCHAR		'0'
#define DOTCHAR			'.'
#define XTRSYMCHAR		'$'
#define VARPRECHAR		'_'

//======================================================================
// check CharSets

inline bool isSignChar(char c) { return (c == PLUSCHAR || c == MINUSCHAR) ; }
inline bool isZeroChar(char c) { return (c == ZEROCHAR) ; }
inline bool isDotChar(char c) { return (c == DOTCHAR) ; }
inline bool isHexPreChar(char c) { return (c == 'x' || c == 'X') ; }
inline bool isOctIntChar(char c) { return ('0' <= c && c <= '7');} 
inline bool isExpSepChar(char c) { return (c == 'e' || c == 'E') ; }

inline bool isSymHeadChar(char c) { return (islower(c) || c == XTRSYMCHAR);}
inline bool isVarHeadChar(char c) { return (isupper(c) || c == VARPRECHAR);}
inline bool isSymBodyChar(char c) { 
   return (isalnum(c) || c == XTRSYMCHAR || c == VARPRECHAR) ;
} 

inline bool isArgBeginChar(char c) { return (c == ARGBEGIN) ; }
inline bool isArgEndChar(char c) { return (c == ARGEND) ; }
inline bool isArgSepChar(char c) { return (c == ARGSEP) ; }


//======================================================================
// OTHER CONSTANTS

#define BUFSIZE 1024 
#define EXTENDMAGNIFY 2

//======================================================================
// class Local Buffer
//----------------------------------------------------------------------
class LocalBuffer {
   public:
      char *buffer ;
      Int len ;
      Int max ;
      bool extendp ;
      bool originalp ;

      inline bool assignBuffer(Int maxlen) {
	 buffer = new char[maxlen] ;
	 originalp = true ;
	 max = maxlen ;
	 return true ;
      } ;

      inline LocalBuffer(Int maxlen = 0,
			 bool exp = false) {
			    originalp = false ;
			    len = 0 ;
			    max = maxlen ;
			    buffer = NULLPTR ;
			    if(max > 0) assignBuffer(max) ;
			    extendp = exp ;
      } ;

      inline LocalBuffer(char* buf,
			 Int maxlen,
			 bool exp = false,
			 bool copyp = false) {
			    originalp = false ;
			    len = 0 ;
			    max = maxlen ;
			    if(copyp) {
			       assignBuffer(max) ;
			       strncpy(buffer,buf,max) ;
			    } else {
			       buffer = buf ;
			    }
			    extendp = exp ;
      } ;

      inline LocalBuffer(LocalBuffer& orgbuf,
			 Int offset = 0,
			 Int l = 0,
			 bool copyp = false) {
			    originalp = false ;
			    len = l ;
			    max = orgbuf.max - offset ;
			    if(copyp) {
			       assignBuffer(max) ;
			       strncpy(buffer,&(orgbuf.buffer[offset]),max) ;
			    } else {
			       buffer = &(orgbuf.buffer[offset]) ;
			    }
			    extendp = orgbuf.extendp ;
      } ;

      inline ~LocalBuffer() {
	 if(originalp) delete buffer ;
      } ;
      inline bool deleteBuffer() {
	 if(originalp) delete buffer ;
	 return originalp ;
      } ;

      inline bool extendBuffer() {
	 char* oldbuf = buffer ;
	 bool origp = originalp ;
	 bool r = assignBuffer(max * EXTENDMAGNIFY) ;
	 strncpy(buffer,oldbuf,len) ;
	 if(origp) deleteBuffer() ;

	 return r ;
      } ;

      bool putChar(char c) ;
      inline bool finish() { return putChar(EOSCHAR) ; }
} ;

//======================================================================
// class ReadStatus
//----------------------------------------------------------------------
enum ReadErrorID {
   REI_None,
   REI_NilData,
   REI_IllegalEOF,
   REI_IllegalNumberFormat,
   REI_IllegalSymbolFormat,
   REI_IllegalStringFormat,
   REI_IllegalTermFormat,
} ;

class ReadStatus {
   public:
      ReadErrorID errorID ;
      Int charCount ;
      Int lineCount ;
  
      inline ReadStatus() {
	 errorID = REI_None ;
	 charCount = 0  ;
	 lineCount = 0 ;
      } ;

      inline void incChar(char c) {
	 charCount++ ;
	 if(c == CRCHAR) lineCount++ ;
      } ;
} ;

//======================================================================
// Utility
//
extern Int skipSpace(istream& istr, ReadStatus& rstatus) ;
extern bool skipComment(istream& istr, ReadStatus& rstatus) ;
extern void skipSpaceComment(istream& istr, ReadStatus& rstatus) ;
extern bool readUntilChar(istream& istr, LocalBuffer& buf, 
			  char endchar, char escapechar,
			  bool discardlastp, ReadStatus& rstatus) ;
extern bool skipLine(istream& istr, ReadStatus& rstatus) ;

inline char getChar(istream& istr, ReadStatus& rstatus) {
   char c ;
/*     This part is for problem of old istream in multithread */
#ifdef xxxTHREADEDxxx
/*	istream bug?? 	
	Following code can not wait input.
	if((c = istr.peek()) != EOF) {
	istr.get(c) ;
	}
	*/
   do {
      c = NULLCHAR ;
      istr.get(c) ;
      if (c == NULLCHAR) { 
	 istr.clear();
	 MSleep(10) ;
      }
   } while (c == NULLCHAR) ;
/*     istream in the new version is MT safe, so, istr.get is fine. */ 
#else
   istr.get(c) ;
#endif
   rstatus.incChar(c) ;
   return c ;
} ;

//======================================================================
// readGData
//----------------------------------------------------------------------
extern GData readSentence(istream& istr, ReadStatus& rstatus) ;
inline GData readSentence(istream& istr) {
   ReadStatus rstatus ;
   return readSentence(istr,rstatus) ;
}
inline GData readSentence(char* str) {
   istrstream istr(str) ;
   return readSentence(istr) ;
}
   

extern GData readGData(istream& istr, ReadStatus& rstatus) ;
inline GData readGData(istream& istr) {
   ReadStatus rstatus ;
   return readGData(istr,rstatus) ;
}
inline GData readGData(char* str) {
   istrstream istr(str) ;
   return readGData(istr) ;
}

extern GData readQuotedGSymbol(istream& istr, ReadStatus& rstatus) ;
extern GData readGString(istream& istr, ReadStatus& rstatus) ;
extern GData readGNumberGSymbol(istream& istr, ReadStatus& rstatus) ;
extern GData readGIntFloatRest(istream& istr, LocalBuffer& buf, 
			       ReadStatus& rstatus);
extern GData readGOctIntRest(istream& istr, LocalBuffer& buf, 
			     ReadStatus& rstatus);
extern GData readGHexIntRest(istream& istr, LocalBuffer& buf, 
			     ReadStatus& rstatus);
extern GData readGSymbolRest(istream& istr, LocalBuffer& buf, 
			     ReadStatus& rstatus);
extern GData readGVariableRest(istream& istr, LocalBuffer& buf, 
			       ReadStatus& rstatus);
extern GData checkArgs(const GData& head, istream& istr, ReadStatus& rstatus);
extern GData readArgs(const GData& head, istream& istr, ReadStatus& rstatus,
		      Int argn) ;

//======================================================================
// output for debug
//----------------------------------------------------------------------
extern void dbp(GData dt) ;
extern void dbd(GData dt) ;

//======================================================================
// Other Util
//----------------------------------------------------------------------
  





////////////////////////////////////////////////////////////////////////
#endif
