// -*- Mode:C++ -*-
//Header:
//File: gc.h
//Author: NODA Itsuki
//Date: 1998/1/22
//EndHeader:

//ModifyHistory:
// 1998/1/22: Create this file
//EndModifyHistory:

//Memo:
//  GCNode: node of link from GCTopNode to GCEntity.
//	GCNode may not be actual entity.
//  GCEntity: actual entity of heap memory.
//  GCTopNode: top node where markup process starts.
//EndMemo:

#ifndef _gc_h_
#define _gc_h_
////////////////////////////////////////////////////////////////////////
#include <stdlib.h>
#include <iostream.h>
#include "base.h"
#include "thread_util.h"

//======================================================================
// GCNode
class GCNode {
   public:
      bool markup() ;
} ;

//======================================================================
// GCEntity
class GCEntity : public GCNode {
   public:
      bool	used ;
      int	instack ;

      inline GCEntity() { used = false ; instack = 0 ; } ;

      inline stackIn() { ++instack ; } ;
      inline stackOut() { --instack ; } ;

      bool markup() ;
} ;

//======================================================================
// GCTopNode
class GCTopNode : public GCNode {
   public:

} ;

//======================================================================
// GCEntityTable
class GCEntityTable {
   public:
      GMutex mutex ;

      inline GCEntityTable() {
      } ;

      inline void Lock() { 
	 mutex.lock() ;
      } ;
      inline void UnLock() { 
	 mutex.unlock() ;
      } ;

      bool sweep(bool flag) ;
} ;

//======================================================================
// GCTopNodeTable
class GCTopNodeTable : public GCTopNode {
   public:
      Int n ;
      Int max ;
      GCTopNode **node ;
#ifdef THREADED
      pthread_mutex_t mutex ;
#endif

      inline GCTopNodeTable(Int size = 0) {
	 max = size ;
	 n = 0 ;
	 if(size > 0) node = new (GCTopNode *)[size] ;
	 else node = NULLPTR ;
#ifdef THREADED
	 pthread_mutex_init(&mutex,NULLPTR) ;
#endif
      } ;

      inline bool pushTopNode(GCTopNode* nd) {
#ifdef THREADED
	 pthread_mutex_lock(&mutex) ;
#endif
	 bool r ;
	 if(n < max) {
	    node[n] = nd ;
	    n++ ;
	    r = true ;
	 } else {
	    r = false ;
	 }
#ifdef THREADED
	 pthread_mutex_unlock(&mutex) ;
#endif
	 return r ;
      } ;
     
      bool markup() ;
} ;





////////////////////////////////////////////////////////////////////////
#endif
