// -*- Mode:C++ -*-
//Header:
//File: gcell.cc
//Author: NODA Itsuki
//Date: 1998/6/5
//EndHeader:

//ModifyHistory:
// 1998/6/5: Create this file
//EndModifyHistory:

#include "gcell.h"

//======================================================================
// GClause Table

//--------------------------------------------------
// New Ring

GClauseRing* GClauseTable::newRing() {
   GClauseRing* r ;
   mutex.lock() ;
   if(trash->isSingleRing()) {
      r = new GClauseRing() ;
   } else {
      r = trash->next ;
      r->removeSelf() ;
   }
   mutex.unlock() ;
   return r ;
}

//--------------------------------------------------
// asserta

GClauseRing* GClauseTable::asserta(const GClause& gclause) {
   GClauseRing* r = newRing() ;
   r->set(gclause) ;
   mutex.lock() ;
   root->addNext(r) ;
   mutex.unlock() ;
   return r ;
} 

//--------------------------------------------------
// assertz
	    
GClauseRing* GClauseTable::assertz(const GClause& gclause) {
   GClauseRing* r = newRing() ;
   r->set(gclause) ;
   mutex.lock() ;
   root->addPrevious(r) ;
   mutex.unlock() ;
   return r ;
}

//--------------------------------------------------
// retract
	    
GClauseRing* GClauseTable::retract(GClauseRing* gclring) {
   mutex.lock() ;
   gclring->removeSelf() ;
   trash->addPrevious(gclring) ;
   mutex.unlock() ;
   return gclring ;
}

//--------------------------------------------------
// describe

void GClauseTable::describe(ostream& o, bool detailp = true) const {
   if(detailp) {
      o << "#GClauseTable[" << this << "]" << endl;
      o << "  table=" ;
      root->describe(o,true) ;
      o << "  # of garbage=" << trash->clength() << endl ;
      o << "----- end GClauseTable" << endl ;
   } else {
      o << "#GClauseTable[" << this 
	<< ":# of defs=" << root->clength() 
	<< ", # of garbage=" << trash->clength() 
	<< "]" ;
   } 
} 

//======================================================================
// GCell

//--------------------------------------------------
// copy gcell stack
UInt GCell::copyGCellStack(GCell* src) {
   gcs_mutex.lock() ;
   gcellStack.reset() ;
   UInt n = src->gcellStack.toppos() ;
   for(UInt i = n ; i > 0  ; i--) {
      gcellStack.push(src->gcellStack.topN(i-1)) ;
   }
   gcs_mutex.unlock() ;
   return n ;
}

//--------------------------------------------------
// check gcell is in gcell stack
bool GCell::isInGCellStack(GCell* target) {
   UInt n = gcellStack.toppos() ;
   for(UInt i = 0 ; i < n ; i++) {
      if(gcellStack.topN(i) == target) 
	 return true ;
   }
   return false ;
}

//--------------------------------------------------
// describe
void GCell::describe(ostream& o, bool detailp = true) const {
   o << "#GCell[" ;

   if(isNullPtr(name)) o << this ;
   else o << *name ;

   o << "]" ;

   if(detailp) {
      o << endl ;
      o << "    Definitions:" ;
      gclauseTable.describe(o,true) ;
      o << "    CellStack:" << gcellStack << endl ;
      for(Int i = 0 ; i < gcellStack.toppos() ; i++) {
	 o << "\t" << *(gcellStack.topN(i)) << endl ;
      }
   } 
}

//--------------------------------------------------
// show GCell Stack

void GCell::showGCellStack(ostream& o, bool newlinep) const {
   if(isNullPtr(name)) o << this ;
   else o << *name ;
   if(!gcellStack.bottomp()) {
      o << "(" ;
      bool firstp = true ;
      for(UInt i = 0 ; i < gcellStack.toppos() ; i++) {
	 if(firstp) firstp = false ; 
	 else o << "," ;

	 gcellStack.topN(i)->showGCellStack(o,false) ;
      }
      o << ")" ;
   }
   if(newlinep) o << endl ;
}

//--------------------------------------------------
// listing

void GCell::listing(ostream& o, bool recursivep, bool detailp) const {
   o << "In " << *this << "." << endl ;

//   o << "  " ; 
//   gclauseTable.root->describe(o,true) ;

   GClauseRing* tail = gclauseTable.root ;
   for(GClauseRing* r = gclauseTable.root->next ;
	r != tail ;
	r = r->next) {
      if(detailp) {
	 o << "  " << r << "\t" << r->entity << endl ;
      } else {
	 o << "  " << r->entity << endl ; ;
      }
    }

   if(recursivep) {
      for(UInt i = 0 ; i < gcellStack.toppos() ; i++) {
	 gcellStack.topN(i)->listing(o,recursivep,detailp) ;
      }
   }
} 
      

//======================================================================
// GCellPointer
   
//--------------------------------------------------
// Output

void GCellPointer::describe(ostream& o, bool detailp) const {
   o << "#GCellPointer[" << gcellstack << "," << n << "]" ;
   if(detailp) {
      o << endl ;
      o << "  GCellStack:" << endl ;
      gcellstack->describe(o, true) ;
   } ;
}

//======================================================================
// GClausePointer

//--------------------------------------------------
// push/pop

void GClausePointer::pushgcp(GCell* tcell) {
   if(0 < tcell->gcellStack.toppos()) {
      gcpstack.push(GCellPointer(&(tcell->gcellStack),0)) ;
   } 
} 

GCell* GClausePointer::popgcell() {
  popgcell_top:
   
   if(gcpstack.toppos() <= 0) return NULLPTR ;

   GCellPointer gcp = gcpstack.pop() ;
   if(gcp.gcellstack->toppos() <= gcp.n) 
      goto popgcell_top ;
	 
   GCell* r = gcp.gcellstack->topN(gcp.n) ;
   gcp.n++ ;
   gcpstack.push(gcp) ;
   
   return r ;
} ;

//--------------------------------------------------
// find GClause in Ring

GClauseRing* GClausePointer::findGClauseInRing(const GTermHeadArity& gtermha) {
   GClauseRing* c ;
   for(c = top->next ; c != topcell->gclauseTable.root ; c = c->next ) {
      if(c->entity.head_arity.eq(gtermha)) {
	 top = c ;
	 return c ;
      }
   }
   return NULLPTR ;
}

//--------------------------------------------------
// find GClause

GClauseRing* GClausePointer::findGClause(const GTermHeadArity& gtermha) {
  findGClause_top:

   GClauseRing* r = findGClauseInRing(gtermha) ;
   if(!isNullPtr(r)) return r ;

   pushgcp(topcell) ;
   GCell* newtopcell = popgcell() ;
   if(isNullPtr(newtopcell)) return NULLPTR ;
	 
   settopcell(newtopcell) ;
   goto findGClause_top ;
}

//--------------------------------------------------
// Output

void GClausePointer::describe(ostream& o, bool detailp) const {
   o << "#GClausePointer[" << this << ":" ;
   if(isNullPtr(topcell)) o << "null" ;
   else o << *topcell ;
   o << "]" ;
   if(detailp && !isNullPtr(this)) {
      o << endl ; 
      o << "   top=" << *top << endl ;
      o << "   topcell=" << *topcell << endl ;
      o << "   gcellpointerstack:" << endl; 
      for(Int i = 0 ; i < gcpstack.toppos() ; i++) {
	 (gcpstack.topN(i)).describe(o,true) ;
      }
   } ;
}

