// -*- Mode:C++ -*-
//Header:
//File: gem.cc
//Title: gaea elementary machine
//Author: NODA Itsuki
//Date: 1998/6/09
//EndHeader:

//ModifyHistory:
// 1998/6/09: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#include "gem.h"

//======================================================================
// OrState
//

//--------------------------------------------------
// output

void OrState::describe(ostream& o, bool detailp) const {
   o << "#OrState[" << this << "]" ;
   if(detailp) {
      o << endl ;
      o << "   AndStateStackPointer=" << andStateStackPointer << endl ;
      o << "   BindingState=" << bindingState << endl ;
      o << endl ;
   }
} ;

//======================================================================
// GEM Gaea Elementary Machine

//--------------------------------------------------
// read/exec/print loop

void Gem::reploop0(istream& istr, ostream& ostr) {
   bool endp = false ;
   while(!endp) {
      ostr << "?- " ; ostr.flush() ;
      GData gdt = readSentence(istr) ;
      if(eq(gdt,EOFSymbol)) { endp = true ; continue ; }
      if(eq(gdt,ErrorSymbol)) { 
	 ostr << "Error: Illegal form." << endl ;
	 continue ; 
      }
      if(isGTerm(gdt)) {
	 if((_TermHead(gdt) == _SymVal(RuleOpSymbol) &&
	     _TermArity(gdt) == 2) ||
	    (_TermHead(gdt) == _SymVal(FactOpSymbol) &&
	     _TermArity(gdt) == 1)) {
	    gcell->assertz(GClause(gdt)) ;
	    continue ;
	 }
      }
      settopgoal(gdt) ;
      bool r = unfoldLoop();
      if(r) {
	 ostr << fetchValue(topGoal,&uiheap) << endl ;
      } else {
	 ostr << "Fail!" << endl ;
      }
   }
}

//--------------------------------------------------
// load loop

void Gem::load(String filename, GCell* dstcell) {
   ifstream istr((char*)filename) ;
   if(istr) { 
      loadloop(istr, dstcell) ;
   } else {
      cerr << "Error: Can not open file \"" << filename << "\"" << endl ;
   }
   istr.close() ;
}

void Gem::loadloop(istream& istr, GCell* dstcell) {
   ReadStatus rstatus ;
   bool endp = false ;
   while(!endp) {
      GData gdt = readSentence(istr,rstatus) ;
      if(eq(gdt,ErrorSymbol) || eq(gdt,EOFSymbol)) {
	 endp = true ;
	 continue ;
      }
      if(isGTerm(gdt)) {
	 if((_TermHead(gdt) == _SymVal(RuleOpSymbol) &&
	     _TermArity(gdt) == 2) ||
	    (_TermHead(gdt) == _SymVal(FactOpSymbol) &&
	     _TermArity(gdt) == 1)) {
	    if(isNullPtr(dstcell)) { 
	       agcell()->assertz(GClause(gdt)) ;
	    } else {
	       dstcell->assertz(GClause(gdt)) ;
	    }
	    continue ;
	 }
      }
      UInt orlevel = orStateStack.toppos() ;
      settopgoal(gdt,NULLPTR,false) ;
      bool r = unfoldLoop();
      orStateStack.setpos(orlevel) ;
   }
}

//--------------------------------------------------
// settopgoal

void Gem::settopgoal(GData goal, Binding* binding, bool resetp) {
   if(resetp) {
      andStateStack.reset() ;
      orStateStack.reset() ;
      uiheap.reset() ;
   }
   if(isNullPtr(binding)) 
      binding = uiheap.newBinding() ;
   
   topGoal = InstGData(goal,binding) ;
   andStateStack.push(AndState(topGoal,orStateStack.toppos())) ;
} ;

//--------------------------------------------------
// load loop

void* UnfoldLoop(void* gem) {
   ((Gem*)gem)->unfoldLoop() ;
   return gem ;
}

bool Gem::unfoldLoop() {
   UnfoldState r ;

   while(true) {
      r = unfold() ;
      if(r == US_fail) {
	 if(orStateStack.bottomp()) {
	    return false ;
	 } else {
	    popOrState() ;
	 }
      } else if (r == US_succeed) {
	 return true ;
      }
      Yield() ;
   } ;
}

//--------------------------------------------------
// unfold

Gem::UnfoldState Gem::unfold() {
   if(andStateStack.bottomp())	return US_succeed ;
   else if(tryDef()) 		return US_continue ;
   else 			return US_fail ;
}

//--------------------------------------------------
// try system def

bool Gem::tryDef() {
   astate = andStateStack.toppointer() ;

   bool r = false ;

   if(tracelevel > TRC_none) {
      switch(tracelevel) {
	 case TRC_normal:
	    cerr << "Call:" 
		 << fetchValue(currentGoal(),(UnifyInfoHeap *)&uiheap) 
		 << endl ; 
	    break ;
	 case TRC_kernel:
	    cerr << endl << "<<<<<<<<<<Enter tryDef<<<<<<<<<<" << endl ;
	    showstatus(cerr) ;
	    break ;
      }
   }

   if(trySystemDef()) {
      r = true ;
   } else {
      if(astate->firsttryp()) {
	 astate->initGClausePointer(gcell) ;
      } ;
      if(tryUserDef()) { r = true ;}
      else {r = false ;}
   }

   if(tracelevel > TRC_none) {
      switch(tracelevel) {
	 case TRC_normal:
	    if(r) cerr << "Succ:" ;
	    else  cerr << "Fail:" ;
	    cerr << fetchValue(currentGoal(),(UnifyInfoHeap *)&uiheap) 
		 << endl ; 
	    break ;
	 case TRC_kernel:
	    cerr << endl << ">>>>>>>>>>Exit tryDef>>>>>>>>>>" << endl ;
	    showstatus(cerr) ;
	    break ;
      }
   }

   return r ;
}

//--------------------------------------------------
// try system def

bool Gem::trySystemDef() {
   GData goal = currentGoal().data ;
   if(isGSymbol(goal)
      && _SymVal(goal)->funcp) {
      return (*(_SymVal(goal)->func))(this) ;
   } 
   if(isGTerm(goal)
      && _TermHead(goal)->funcp) {
      return (*(_TermHead(goal)->func))(this) ;
   } 
   return false ;
}

//--------------------------------------------------
// try user def


bool Gem::tryUserDef() {
   Binding* newbinding = uiheap.newBinding() ;
   OrState* ostate = pushOrState() ;
   BindingState* bstate = &(ostate->bindingState) ;

   GClauseRing* gcr ;
   while(true) {
      gcr = gclausePointer()->findGClause(GTermHeadArity(currentGoal().data)) ;
      if(isNullPtr(gcr)) { 
	 popOrState() ;
	 return false ;
      } 

      newbinding->init() ;
      if(unify(currentGoal(),InstGData(gcr->entity.head,newbinding),
	       &uiheap)) {
	 popAndState() ;
	 pushAndState(InstGData(gcr->entity.body,newbinding),
		      orStateStack.toppos()) ;
	 return true ;
      } 
      rewindBindingState(bstate) ;
   }
}

//--------------------------------------------------
// Describe
      
void Gem::describe(ostream& o, bool detailp = true) const {
   o << "#GEM[" << *name << "]"  ;
   if(detailp) {
      o << endl ;
      o << "  pid=" << pid << endl ;
      o << "  env=" ; 
      gcell->describe(o,true) ;
      o << "  top goal=" << topGoal << endl ;

      if(isNullPtr(astate)) { 
	 o << "  current goal= unbounded" << endl ;
	 o << "  gclausePointer= unbounded" << endl ;
      } else {
	 o << "  current goal=" ;
	 o << fetchValue(currentGoal(),(UnifyInfoHeap *)&uiheap) << endl ; 
	 o << "  gclausePointer=" ;
	 o << *gclausePointer() << endl; 
      }
      
      o << "  and state stack=" ;
      andStateStack.describe(o,true) ;
      o << "  or state stack=" ;
      orStateStack.describe(o,true) ;
      o << "  uiheap=" ;
      uiheap.describe(o,true) ;
      o << endl ;
   }
}

//--------------------------------------------------
// show gem status

void Gem::showstatus(ostream& o) const {
   o << "#GEM[" << *name << "]" << endl ;

   o << "  Env=" ;
   gcell->showGCellStack(o,true) ;
   
   o << "  TGoal=" << topGoal << endl ;
   
   if(!isNullPtr(astate)) {
      o << "  CGoal=" ;
      o << fetchValue(currentGoal(),(UnifyInfoHeap *)&uiheap) << endl ; 
      o << "  GCP=" << *gclausePointer() << endl ;
   } ;
   
   o << "  AndStack=" ; 
   andStateStack.describe(o,true) ;
   o << endl ;

   o << "  OrStack=" ;
   orStateStack.describe(o,true) ;
   o << "============================================================" << endl;
}


//--------------------------------------------------
// for debug
      
void dbdGem(Gem* gem) { gem->describe(cerr, true) ; } 


