%======================================================================
% builtin-functions
%======================================================================

$fact(eq(X,X)).

$rule(listing,$and($cut,$listing)).
$rule(listing(Cell),$and($cut,$listing(Cell))).

$rule(halt,$halt).

$rule(load(Filename),$load(Filename)).

$rule(fork(Goal,PID),$fork(Goal,PID)).
$rule(pid(PID),$pid(PID)).

$rule(is(Var,Form),$is(Var,Form)).

% push/pop cause serious change to gclausepointer, 
% so it should be built-in pred

$rule(push(Cell),$and($cut,$push(Cell))).
$rule(push(Cell,Dst),$and($cut,$push(Cell,Dst))).
$rule(pop(Cell),$pop(Cell)).
$rule(pop(Cell,Src),$pop(Cell,Src)).

$rule(repush(Cell,Dst),
	$and($in_env(Cell,Dst),$remove_cell(Cell,Dst),$push(Cell,Dst),$cut)).
$rule(repush(Cell,Dst),
	$push(Cell,Dst)).
$rule(needs(Cell,Dst),
	$and($in_env(Cell,Dst),$cut)).
$rule(needs(Cell,Dst),
	$push(Cell,Dst)).
$rule(remove(Cell,Dst),
	$and($cut,$remove_cell(Cell,Dst))).
$rule(remove_a_cell(Cell,Dst),
	$and($cut,$remove_a_cell(Cell,Dst))).
$rule(fremove(Cell,Dst),
	$and($in_env(Cell,Dst),$remove_cell(Cell,Dst),$cut)).
$rule(fremove(Cell,Dst),
	$cut).
$rule(fremove_a_cell(Cell,Dst),
	$and($in_env(Cell,Dst),$remove_a_cell(Cell,Dst),$cut)).
$rule(fremove_a_cell(Cell,Dst),
	$cut).

$rule(newcell(Cell),$newcell(Cell)).

$rule(pushsecond(NewCell),
	$and($cut,pop(Top),push(NewCell),push(Top))).
$rule(pushsecond(NewCell,Dst),
	$and($cut,pop(Top,Dst),push(NewCell,Dst),push(Top,Dst))).

$rule(gencell(NewCell,Seed),
	$and($cut,$gensym(NewCell,Seed),newcell(NewCell))).

$rule(assert(Fact),asserta(Fact)).
$rule(assert_in(Fact,Dst),asserta_in(Fact,Dst)).

$rule(asserta(Fact),$asserta_fact(Fact)).
$rule(asserta_in(Fact,Dst),$asserta_fact(Fact,Dst)).

$rule(assertz(Fact),$assertz_fact(Fact)).
$rule(assertz_in(Fact,Dst),$assertz_fact(Fact,Dst)).

$rule(environment(Env),$environment(Env)).

$rule(sleep(forever),$and($cut,$sleep_forever)).
$rule(sleep(Sec),$sleep(Sec)).

%======================================================================
% control
%======================================================================

$fact(true).

$rule(repeat,$repeat).
$rule(dotimes(I,M),dotimes(I,0,M)).
$rule(dotimes(I,C,M),
	$and($comp_ge(C,M),$cut,$fail)).
$fact(dotimes(C,C,M)).
$rule(dotimes(I,P,M),
	$and($inc(P,N),$cut,dotimes(I,N,M))).

%======================================================================
% list operations
%======================================================================

$fact(member(Item,$(Item,_))).
$rule(member(Item,$(_,Rest)),
	member(Item,Rest)).

$fact(append($,Tail,Tail)).
$rule(append($(Head,Body),Tail,$(Head,Rest)),
	append(Body,Tail,Rest)).

$rule(reverse(List,Rev),reverse(List,$,Rev)).
$fact(reverse($,Rev,Rev)).
$rule(reverse($(Head,Rest),Stack,Rev),
	reverse(Rest,$(Head,Stack),Rev)).

$fact(length($,0)).
$rule(length($(_,Rest),Len),
	$and($cut,length(Rest,Len1),$inc(Len1,Len))).


