/*
 *	Header:
 *	File: position.c
 *	Author: Noda Itsuki
 *	Date: 1997/04/1
 *	EndHeader:
 */

#include <stdio.h>
#include <math.h>

#include "libsclient.h"

#define Sin(x) sin(Deg2Rad(x))
#define Cos(x) cos(Deg2Rad(x))

Bool estimate_current_pos(SeeInfo *sinf, Side side, PosState *pstate)
{
  int i ;
  int out_of_field = FALSE ;
  double dist = 2.0*PITCH_LENGTH ;
  double line_dir, obj_dir ;
  double tmp ;
  SeenObj *line = (SeenObj *)NULL, *flag = (SeenObj *)NULL ;
  SeenObj *goal = (SeenObj *)NULL, *obj ;
  Pos obj_pos ;

  for(i = 0; i < sinf->n_obj; i++) {
    if (sinf->obj[i].type == OT_Flag) {
      if (!sinf->obj[i].inViewConeP)
	continue ;
      if (Abs(dist) > Abs(sinf->obj[i].dist)) {
	flag = &(sinf->obj[i]) ;
	dist = sinf->obj[i].dist ;
      }
    }
    else if (sinf->obj[i].type == OT_Line) {
      line = &(sinf->obj[i]) ;
      if (i != sinf->n_obj - 1 && sinf->obj[i+1].type == OT_Line)
	out_of_field = TRUE ;
      break ;
    }
    else if (sinf->obj[i].type == OT_Goal)
      goal = &(sinf->obj[i]) ;
  }

  if (line == (SeenObj *)NULL)
    return FALSE ;

  if (flag != (SeenObj *)NULL) {
    obj = flag ;
    for(i=0; ; i++) {
      if (Flags[i].name == FP_Unknown)
	return FALSE ;
      if (Flags[i].name == flag->id.fpos) {
	obj_pos.x = Flags[i].pos.x ;
	obj_pos.y = Flags[i].pos.y ;
	break ;
      }
    }
  }
  else if (goal != (SeenObj *)NULL) {
    obj = goal ;
    if (goal->id.side == S_LEFT)
      obj_pos.x = -PITCH_LENGTH/2.0 ;
    else if (goal->id.side == S_RIGHT)
      obj_pos.x = PITCH_LENGTH/2.0 ;
    else
      return FALSE ;
    obj_pos.y = 0.0 ;
  }
  else
    return FALSE ;

  line_dir = Deg2Rad(line->dir) ;

  switch (line->id.lpos) {
  case LP_T :
    if (out_of_field)  {
      pstate->dir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
      pstate->pos.y = -PITCH_WIDTH/2.0 - Abs(line->dist * sin(line_dir)) ;
    }
    else {
      pstate->dir = -line->dir - ((line->dir < 0) ? 180.0 : 0.0) ;
      pstate->pos.y = -PITCH_WIDTH/2.0 + Abs(line->dist * sin(line_dir)) ;
    }

    if ((tmp = Pow(obj->dist) - Pow(obj_pos.y - pstate->pos.y)) < 0.0)
      tmp *= -1.0 ;

    obj_dir = obj->dir + pstate->dir ;
    if (obj_dir > 180.0)
      obj_dir -= 360.0 ;
    else if (obj_dir < -180.0)
      obj_dir += 360.0 ;

    if (Abs(obj_dir) < 90.0)
      pstate->pos.x = obj_pos.x - sqrt(tmp) ;
    else
      pstate->pos.x = obj_pos.x + sqrt(tmp) ;
    break ;

  case LP_B :
    if (out_of_field) {
      pstate->dir = -line->dir - ((line->dir < 0) ? 180.0 : 0.0) ;
      pstate->pos.y = PITCH_WIDTH/2.0 + Abs(line->dist * sin(line_dir)) ;
    }
    else {
      pstate->dir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
      pstate->pos.y = PITCH_WIDTH/2.0 - Abs(line->dist * sin(line_dir)) ;
    }

    if ((tmp = Pow(obj->dist) - Pow(obj_pos.y - pstate->pos.y)) < 0.0)
      tmp *= -1.0 ;

    obj_dir = obj->dir + pstate->dir ;
    if (obj_dir > 180.0)
      obj_dir -= 360.0 ;
    else if (obj_dir < -180.0)
      obj_dir += 360.0 ;

    if (Abs(obj_dir) < 90.0)
      pstate->pos.x = obj_pos.x - sqrt(tmp) ;
    else
      pstate->pos.x = obj_pos.x + sqrt(tmp) ;
    break ;

  case LP_L :
    if (out_of_field) {
      pstate->dir = -line->dir - ((line->dir > 0) ? -90.0 : 90.0) ;
      pstate->pos.x = -PITCH_LENGTH/2.0 - Abs(line->dist * sin(line_dir)) ;
    }
    else {
      pstate->dir = -line->dir - ((line->dir < 0) ? -90.0 : 90.0) ;
      pstate->pos.x = -PITCH_LENGTH/2.0 + Abs(line->dist * sin(line_dir)) ;
    }

    if ((tmp = Pow(obj->dist) - Pow(obj_pos.x - pstate->pos.x)) < 0.0)
      tmp *= -1.0 ;

    obj_dir = obj->dir + pstate->dir ;
    if (obj_dir > 180.0)
      obj_dir -= 360.0 ;
    else if (obj_dir < -180.0)
      obj_dir += 360.0 ;

    if (obj_dir > 0.0)
      pstate->pos.y = obj_pos.y - sqrt(tmp) ;
    else
      pstate->pos.y = obj_pos.y + sqrt(tmp) ;
    break ;

  case LP_R :
    if (out_of_field) {
      pstate->dir = -line->dir - ((line->dir < 0) ? -90.0 : 90.0) ;
      pstate->pos.x = PITCH_LENGTH/2.0 + Abs(line->dist * sin(line_dir)) ;
    }
    else {
      pstate->dir = -line->dir - ((line->dir > 0) ? -90.0 : 90.0) ;
      pstate->pos.x = PITCH_LENGTH/2.0 - Abs(line->dist * sin(line_dir)) ;
    }

    if ((tmp = Pow(obj->dist) - Pow(obj_pos.x - pstate->pos.x)) < 0.0)
      tmp *= -1.0 ;

    obj_dir = obj->dir + pstate->dir ;
    if (obj_dir > 180.0)
      obj_dir -= 360.0 ;
    else if (obj_dir < -180.0)
      obj_dir += 360.0 ;

    if (obj_dir > 0.0)
      pstate->pos.y = obj_pos.y - sqrt(tmp) ;
    else
      pstate->pos.y = obj_pos.y + sqrt(tmp) ;
    break ;

  case LP_Unknown :
    return FALSE ;
  }

  if (side == S_RIGHT) {
    pstate->dir += ((pstate->dir > 0) ? -180.0 : 180.0) ;
    pstate->pos.x *= -1.0 ;
    pstate->pos.y *= -1.0 ;
  }
		
  return TRUE ;
}

Bool estimate_current_pos2(SeeInfo *sinf, Side side, PosState *pstate)
{
     int i ;
     int out_of_field = FALSE ;
     SeenObj *line = (SeenObj *)NULL ;
     double sum_x = 0.0, sum_y = 0.0 , n_x = 0.0, n_y = 0.0;

     for(i = 0; i < sinf->n_obj; i++) {
	  if (sinf->obj[i].type == OT_Line) {
	       if(line == NULL) {
		    line = &(sinf->obj[i]) ;
	       } else {
		    out_of_field = TRUE ;
	       }
	  }
     }
     if (line == (SeenObj *)NULL)
	  return FALSE ;

     switch (line->id.lpos) {
     case LP_T :
	  if (out_of_field)  {
	       double ey, w ;
	       pstate->dir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
	       ey = -PITCH_WIDTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_y += ey * w ; n_y += w ;
	  } else {
	       double ey, w ;
	       pstate->dir = -line->dir - ((line->dir < 0) ? 180.0 : 0.0) ;
	       ey = -PITCH_WIDTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_y += ey * w ; n_y += w ;
	  }
	  break ;
     case LP_B :
	  if (out_of_field) {
	       double ey, w ;
	       pstate->dir = -line->dir - ((line->dir < 0) ? 180.0 : 0.0) ;
	       ey = PITCH_WIDTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_y += ey * w ; n_y += w ;
	  } else {
	       double ey, w ;
	       pstate->dir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
	       ey = PITCH_WIDTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_y += ey * w ; n_y += w ;
	  }
	  break ;
     case LP_L :
	  if (out_of_field) {
	       double ex, w ;
	       pstate->dir = -line->dir - ((line->dir > 0) ? -90.0 : 90.0) ;
	       ex = -PITCH_LENGTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_x += ex * w ; n_x += w ;
	  } else {
	       double ex, w ;
	       pstate->dir = -line->dir - ((line->dir < 0) ? -90.0 : 90.0) ;
	       ex = -PITCH_LENGTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_x += ex * w ; n_x += w ;
	  } 
	  break ;
     case LP_R :
	  if (out_of_field) {
	       double ex, w ;
	       pstate->dir = -line->dir - ((line->dir < 0) ? -90.0 : 90.0) ;
	       ex = PITCH_LENGTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_x += ex * w ; n_x += w ;
	  } else {
	       double ex, w ;
	       pstate->dir = -line->dir - ((line->dir > 0) ? -90.0 : 90.0) ;
	       ex = PITCH_LENGTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_x += ex * w ; n_x += w ;
	  }
	  break ;
     case LP_Unknown :
	  return FALSE ;
     }

     for(i = 0 ; i < sinf->n_obj ; i++) {
	  Pos obj_pos ;
	  double obj_dir, obj_dist ;
	  double ex, ey, w ;
	  SeenObj *obj ;
	  obj = &(sinf->obj[i]) ;

	  if(!(obj->inViewConeP)) continue ;
	  if(obj->type == OT_Flag) {
	       obj_pos.x = Flags[obj->id.fpos].pos.x ;
	       obj_pos.y = Flags[obj->id.fpos].pos.y ;
	  } else if(obj->type == OT_Goal) {
	       if(obj->id.side == S_LEFT) 
		    obj_pos.x = -PITCH_LENGTH/2.0 ;
	       else 
		    obj_pos.x =  PITCH_LENGTH/2.0 ;
	       obj_pos.y = 0.0 ;
	  } else {
	       continue ;
	  }

	  obj_dir = obj->dir + pstate->dir ;
	  obj_dist = obj->dist ;
	  ex = obj_pos.x - obj_dist * Cos(obj_dir) ;
	  ey = obj_pos.y - obj_dist * Sin(obj_dir) ;
	  w = 1.0 / Abs(obj_dist) ;
	  sum_x += w * ex ; n_x += w ;
	  sum_y += w * ey ; n_y += w ;
     }

     if(n_x + n_y <= 0.0) return FALSE ;

     pstate->pos.x = sum_x / n_x ;
     pstate->pos.y = sum_y / n_y ; 

     if (side == S_RIGHT) {
	  pstate->dir += ((pstate->dir > 0) ? -180.0 : 180.0) ;
	  pstate->pos.x *= -1.0 ;
	  pstate->pos.y *= -1.0 ;
     }
     return TRUE ;
}

