// -*- Mode:C++ -*-
//Header:
//File: thread_util.cc
//Author: NODA Itsuki
//Date: 1998/2/6
//EndHeader:

#include "thread_util.h"

//======================================================================
// SLEEP
//

//==================================================
// USleep(sec,usec)

#if defined(SELECTUSLEEP) || !defined(THREADED)
//--------------------------------------------------
// 		usleep using select
void USleep(const UInt sec, const UInt usec) {
   struct timeval time_out ;

   time_out.tv_sec = sec ;
   time_out.tv_usec = usec ;

   select(0,(fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL, &time_out) ;
}
#else
//--------------------------------------------------
// 		usleep using cond_timedwait
static pthread_cond_t _NeverAwakeCV = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t _NeverAwakeMTX = PTHREAD_MUTEX_INITIALIZER; 

void USleep(const UInt sec, const UInt usec) {
   struct timeval ctime ;

   gettimeofday(&ctime,NULL) ;

   struct timespec etime ;

   UInt eusec = usec + ctime.tv_usec ;
   etime.tv_nsec = (eusec % 1000000) * 1000 ;
   etime.tv_sec = sec + ctime.tv_sec + (eusec / 1000000) ;

   pthread_mutex_lock(&_NeverAwakeMTX) ;
   pthread_cond_timedwait(&_NeverAwakeCV,&_NeverAwakeMTX,&etime) ;
   pthread_mutex_unlock(&_NeverAwakeMTX) ;
}
#endif

//==================================================
// usleep(usec)

void USleep(const UInt usec) {
   UInt sec = usec / 1000000 ;
   UInt ausec = usec % 1000000 ;
  
   USleep(sec,ausec) ;
}

//==================================================
// MSleep(sec,msec)

void MSleep(const UInt sec, const UInt msec) {
   UInt asec = sec + msec/1000 ;
   UInt ausec = (msec%1000) * 1000 ;

   USleep(asec,ausec) ;
}

//==================================================
// MSleep(msec)

void MSleep(const UInt msec) {
   UInt sec = msec/1000 ;
   UInt usec = (msec%1000) * 1000 ;

   USleep(sec,usec) ;
}

//==================================================
// Sleep(sec)

void Sleep(const UInt sec) {
   USleep(sec,0) ;
}

void FSleep(const Float sec) {
   UInt isec = (UInt)sec ;
   UInt iusec = (UInt)(sec * 1000000) ;

   USleep(isec,iusec) ;
}

//==================================================
// sleep forever

void SleepForever() {
   pthread_mutex_lock(&_NeverAwakeMTX) ;
   pthread_cond_wait(&_NeverAwakeCV,&_NeverAwakeMTX) ;
   pthread_mutex_unlock(&_NeverAwakeMTX) ;
}


