// -*- Mode:C++ -*-
//Header:
//File: unify.cc
//Author: NODA Itsuki
//Date: 1998/3/28
//EndHeader:

//ModifyHistory:
// 1998/3/28: Create this file
//EndModifyHistory:

//======================================================================
// Include Files
//
#include "unify.h"

//======================================================================
// InstGData
//
ostream& operator<< (ostream& o, const InstGData& dt) {
   o << "#InstGData[" << dt.data ;
   o << "|" << dt.binding ;
   o << "]" ;
   return o ;
} ;

//======================================================================
// Bind
//
ostream& operator<< (ostream& o, const Bind& bind) {
   o << "#Bind[" << bind.variable ;
   o << " = " << bind.value ;
   o << "]" ;
   return o ;
} ;

//======================================================================
// New Var Entry
//
ostream& operator<< (ostream& o, const NewVarEntry& nventry) {
   o << "#NewVar[" << nventry.variable ;
   o << " -> " << nventry.newvariable ;
   o << "]" ;
   return o ;
} ;

GData newVariable(const InstGData& var, 
		  NewVarList* nvlist,
		  UnifyInfoHeap* uiheap) 
{
   for(NewVarList* nv = nvlist->next ; !(nv->endp()) ; 
					nv = nv->next ) {
      if((eq(nv->element.variable.data, var.data))&&
	 (nv->element.variable.binding == var.binding)) {
	 return nv->element.newvariable ;
      }
   }
   GData newvar = genGVariable(var.data) ;	
   NewVarList* nv = 
      uiheap->pushNewVarList(NewVarEntry(var,newvar),nvlist->next) ;
   nvlist->next = nv ;
   return newvar ;
} ;

//======================================================================
// Binding
//
Bind* Binding::findBind(const GData var) {
   BindList* blist ;
   for(blist = bindlist ; !(blist->endp()) ; blist = blist->next ) {
      if(eq(blist->element.variable, var)) {
	 return &(blist->element) ;
      }
   }
   return NULLPTR ;
} ;

InstGData Binding::findBindValue(const GData var) {
   Bind* r = findBind(var) ;
   if(r == NULLPTR)
      return InstGData() ;
   else
      return r->value ;
} ;

BindList* Binding::pushBindList(GData var, InstGData val, 
				UnifyInfoHeap* uiheap) 
{
   BindList* newblist = 
      uiheap->pushBindList(Bind(var,val), bindlist) ;
   return bindlist = newblist ;
} ;

ostream& operator<< (ostream& o, const Binding& binding) {
   o << "#Binding[" << (void*)&binding ;
   o << ":" << binding.bindlist->length() ;
   o << ":" << binding.bindlist ;
//   o << "," << binding.newvartable ;
   o << "]" ;
   return o ;
} ;
ostream& operator<< (ostream& o, const Binding* binding) {
   o << "#Binding[" << (void *)binding ;
   if(!isNullPtr(binding)) {
      o << ":" << binding->bindlist->length() ;
   }
   o << "]" ;
   return o ;
} ;



//======================================================================
// isAssigned
//

Bind* isAssigned(GData var, Binding* binding) { 
   return binding->findBind(var) ;
} ;

InstGData isAssignedIGD(GData var, Binding* binding) {
   return binding->findBindValue(var) ;
} ;

bool isGround(GData var, Binding* binding) {
   if (isGVariable(var)) {
      Bind* bind = isAssigned(var, binding) ;
      if(!isNullPtr(bind)) 
	 return isGround(bind->value.data, bind->value.binding) ;
      else 
	 return false ;
   } else {
      return true ;
   }
} ; 

//======================================================================
// Safty Unify (rewind if fail)
//

bool Unify(InstGData x, InstGData y, UnifyInfoHeap* uiheap) {
   BindingState2 bstate(x.binding,y.binding,uiheap) ;
   bool r = unify(x,y,uiheap) ;
   if(r) {
      return true ;
   } else {
      rewindBindingState2(bstate) ;
      return false ;
   }
}

//======================================================================
// Unify
//

bool unify(InstGData x, InstGData y, UnifyInfoHeap* uiheap) {
  _unify_:
   if       (eq(x.data, y.data) && (x.binding == y.binding)) {
      return true ;
   } else if(isGVariable(x.data)) {
      Bind* as = isAssigned(x) ;
      if(!isNullPtr(as)) {
	 x = as->value ;
	 goto _unify_ ;
      } else {
	 linkBind(x,y,uiheap) ;
	 return true ;
      }
   } else if(isGVariable(y.data)) {
      Bind* as = isAssigned(y) ;
      if(!isNullPtr(as)) {
	 y = as->value ;
	 goto _unify_ ;
      } else {
	 linkBind(y,x,uiheap) ;
	 return true ;
      }
   } else if(!isGTerm(x.data) || !isGTerm(y.data)) {
      return eq(x.data, y.data) ;
   } else if(false) {
      /* GGTerm operation here */
   } else {
      return unifyGTerm(x,y,uiheap) ;
   }
} ;

bool unifyGTerm(InstGData x, InstGData y, UnifyInfoHeap* uiheap) {
   Term* xterm = x.data.val.term ;
   Term* yterm = y.data.val.term ;
   if(xterm->head != yterm->head ||
      xterm->arity != yterm->arity) 
      return false ;
   Binding* xbinding = x.binding ;
   Binding* ybinding = y.binding ;
   for(Int i = 0 ; i < xterm->arity ; i++) {
      if(!unify(InstGData(xterm->argv(i),xbinding),
		InstGData(yterm->argv(i),ybinding),
		uiheap))
	 return false ;
   }
   return true ;
}

void linkBind(InstGData var, InstGData val, UnifyInfoHeap* uiheap) {
   var.binding->pushBindList(var.data, val, uiheap) ;
} ;

//======================================================================
// fetch
//

InstGData fetch(InstGData idt) {
   GData dt = idt.data ;
   Binding* binding = idt.binding ;
     
   if(isGVariable(dt)) {
      Bind* bind = binding->findBind(dt) ;
      if(bind == NULLPTR) {
	 return idt ;
      } else {
	 return fetch(bind->value) ;
      }
   } else {
      return idt ;
   }
}

//======================================================================
// fetch-value
//

GData fetchValue(const InstGData& idt, UnifyInfoHeap* uiheap) {
   uiheap->newvarListHeap.reset() ;
   NewVarList* nvlist = uiheap->pushNewVarList(NewVarEntry(), NULLPTR) ;
   return fetchValueBody(idt, uiheap, nvlist) ;
} ;

GData fetchValueBody(const InstGData& idt, UnifyInfoHeap* uiheap, 
		     NewVarList* nvlist) 
{
   GData dt = idt.data ;
   Binding* binding = idt.binding ;
   if        (isGVariable(dt)) {
      Bind* bind = isAssigned(dt,binding) ;
      if(NULLPTR == bind) {
	 return newVariable(idt,nvlist,uiheap) ;
      } else {
	 return fetchValueBody(bind->value, uiheap, nvlist) ;
      }
   } else if (isGAtom(dt)) {
      return dt ;
   } else { /* GTerm */
      return fetchValueGTerm(idt, uiheap, nvlist) ;
   }
} ;

GData fetchValueGTerm(const InstGData& idt, UnifyInfoHeap* uiheap, 
		      NewVarList* nvlist) 
{
   GData dt = idt.data ;
   Binding* binding = idt.binding ;

   GData newdt = GTerm(dt.val.term->head, dt.val.term->arity) ;
   for(UInt i = 0 ; i < dt.val.term->arity ; i++) {
      newdt.val.term->argv(i) = 
	 fetchValueBody(InstGData(dt.val.term->argv(i),binding),
			uiheap, nvlist) ;
   }
   return newdt ;
}
			      
